#	Copyright (C) 2017 Battelle Memorial Institute
import json;
import sys;
import numpy as np;
import matplotlib as mpl;
import matplotlib.pyplot as plt;
import xarray as xr

class TransmissionMetricsProcessor:
    casename = ""
    casepath =""
    bus_metrics = None
    gen_metrics = None
    def __init__(self,case_name="",case_path=""):
        self.casename = case_name
        self.casepath = case_path

    def loadAllMetricsFromJSONFiles(self,case_name="",case_path=""):
        if(len(case_name)>0):
            self.casename = case_name
        if(len(case_path)>0):
            self.casepath = case_path
        
        lp = open(self.casepath + self.casename + "_m_dict.json").read()
        dict = json.loads(lp)
        baseMVA = dict['baseMVA']
        ampFactor = dict['ampFactor']
        gen_keys = list(dict['generators'].keys())
        gen_keys.sort()
        bus_keys = list(dict['fncsBuses'].keys())
        bus_keys.sort()
        print("\n\nFile", self.casename, "has baseMVA", baseMVA, "with GLD load scaling =", ampFactor)
        # print("\nGenerator Dictionary:")
        # print("Unit Bus Type Pnom Pmax Costs[Start Stop C2 C1 C0]")
        # for key in gen_keys:
        #     row = dict['generators'][key]
        #     print(key, row['bus'], row['bustype'], row['Pnom'], row['Pmax'], "[", row['StartupCost'],
        #           row['ShutdownCost'], row['c2'], row['c1'], row['c0'], "]")

        print("\nFNCS Bus Dictionary:")
        print("Bus Pnom Qnom [GridLAB-D Substations]")
        for key in bus_keys:
            row = dict['fncsBuses'][key]
            print(key, row['Pnom'], row['Qnom'], row['GLDsubstations'])

        # read the bus metrics file
        lp_b = open(self.casepath + "bus_" + self.casename + "_metrics.json").read()
        lst_b = json.loads(lp_b)
        # print("\nBus Metrics data starting", lst_b['StartTime'])

        # make a sorted list of the times, and NumPy array of times in hours
        lst_b.pop('StartTime')
        # lst_b.pop('System base MVA')
        # lst_b.pop('Number of buses')
        # lst_b.pop('Number of generators')
        # lst_b.pop('Network name')
        meta_b = lst_b.pop('Metadata')
        times = list(map(int, list(lst_b.keys())))
        times.sort()
        print("There are", len(times), "sample times at", times[1] - times[0], "second intervals")

        hrs = np.array(times, dtype=np.float)
        denom = 3600.0
        hrs /= denom
        time_interval_hours = (times[1] - times[0]) / denom

        # parse the metadata for things of specific interest
        print("\nBus Metadata [Variable Index Units] for", len(lst_b[str(times[0])]), "objects")
        for key, val in meta_b.items():
            #    print (key, val['index'], val['units'])
            if key == 'LMP_P':
                LMP_P_IDX = val['index']
                LMP_P_UNITS = val['units']
            elif key == 'LMP_Q':
                LMP_Q_IDX = val['index']
                LMP_Q_UNITS = val['units']
            elif key == 'PD':
                PD_IDX = val['index']
                PD_UNITS = val['units']
            elif key == 'PQ':
                QD_IDX = val['index']
                QD_UNITS = val['units']
            elif key == 'Vang':
                VANG_IDX = val['index']
                VANG_UNITS = val['units']
            elif key == 'Vmag':
                VMAG_IDX = val['index']
                VMAG_UNITS = val['units']
            elif key == 'Vmax':
                VMAX_IDX = val['index']
                VMAX_UNITS = val['units']
            elif key == 'Vmin':
                VMIN_IDX = val['index']
                VMIN_UNITS = val['units']

        # create a NumPy array of all bus metrics
        data_b = np.empty(shape=(len(bus_keys), len(times), len(lst_b[str(times[0])][bus_keys[0]])), dtype=np.float)
        print("\nConstructed", data_b.shape, "NumPy array for Buses")
        j = 0
        for key in bus_keys:
            i = 0
            for t in times:
                ary = lst_b[str(t)][bus_keys[j]]
                data_b[j, i, :] = ary
                i = i + 1
            j = j + 1

        # display some averages
        # print("Average real power LMP =", data_b[0, :, LMP_P_IDX].mean(), LMP_P_UNITS)
        # print("Maximum bus voltage =", data_b[0, :, VMAG_IDX].max(), VMAG_UNITS)
        # print("Minimum bus voltage =", data_b[0, :, VMIN_IDX].max(), VMIN_UNITS)

        # read the generator metrics file
        lp_g = open(self.casepath + "gen_" + self.casename + "_metrics.json").read()
        lst_g = json.loads(lp_g)
        # print("\nGenerator Metrics data starting", lst_g['StartTime'])

        # make a sorted list of the times, and NumPy array of times in hours
        lst_g.pop('StartTime')
        meta_g = lst_g.pop('Metadata')
        # print("\nGenerator Metadata [Variable Index Units] for", len(lst_g[str(times[0])]), "objects")
        for key, val in meta_g.items():
            # print(key, val['index'], val['units'])
            if key == 'Pgen':
                PGEN_IDX = val['index']
                PGEN_UNITS = val['units']
            elif key == 'Qgen':
                QGEN_IDX = val['index']
                QGEN_UNITS = val['units']

        # create a NumPy array of all bus metrics
        data_g = np.empty(shape=(len(gen_keys), len(times), len(lst_g[str(times[0])][gen_keys[0]])), dtype=np.float)
        print("\nConstructed", data_g.shape, "NumPy array for Generators")
        j = 0
        for key in gen_keys:
            i = 0
            for t in times:
                ary = lst_g[str(t)][gen_keys[j]]
                data_g[j, i, :] = ary
                i = i + 1
            j = j + 1

        ## transform to a xarray dataset


        self.bus_metrics = xr.Dataset({'LMP_P': (['busNum', 'time'], data_b[:, :, LMP_P_IDX]),
                                     'PD': (['busNum', 'time'], data_b[:, :, PD_IDX]),
                                     'QD': (['busNum', 'time'], data_b[:, :, QD_IDX]),
                                     'VMAG': (['busNum', 'time'], data_b[:, :, VMAG_IDX]),
                                     },
                                    coords={'busNum': list(map(int, bus_keys)),
                                            'time': hrs})  # or hrs
        ## TODO add the bus-feeder and bus substation relationship to the attrs
        #bus_metrics.attrs["bus_substation_map"] = {7: 'SUBSTATION7'}
        self.bus_metrics.attrs["bus_info_dict"] = dict['fncsBuses']

        # processing the generation part metrics

        # TODO The emission rate info could be specified in a JSON file
        # generation emission rate table (lb/MWh)
        # Combination of the tables 11.3 and 11. 5 in the last years' report
        #                         CO2                SOX        NOX
        # coal                   205.57*10.09     0.1*10.09     0.06*10.09
        # natrual gas (CC)       117.08*7.67      0.001*7.67    0.0075*7.67
        # natrual gas (CT)       117.08*11.37     0.001*11.37   0.0075*11.37

        gen_emission_rate = {'coal': [205.57 * 10.09, 0.1 * 10.09, 0.06 * 10.09],
                             'gas_combinedcycle': [117.08 * 7.67, 0.001 * 7.67, 0.0075 * 7.67],
                             'gas_singlecycle': [117.08 * 11.37, 0.001 * 11.37, 0.0075 * 11.37]}
        #print('gen_emission_rate')

        gen_cost = np.empty(shape=(len(gen_keys), len(times)), dtype=np.float)
        gen_revenue = np.empty(shape=(len(gen_keys), len(times)), dtype=np.float)
        gen_emission_co2 = np.empty(shape=(len(gen_keys), len(times)), dtype=np.float)
        gen_emission_sox = np.empty(shape=(len(gen_keys), len(times)), dtype=np.float)
        gen_emission_nox = np.empty(shape=(len(gen_keys), len(times)), dtype=np.float)

        co2_emission_rate = np.empty(shape=(len(gen_keys)), dtype=np.float)
        sox_emission_rate = np.empty(shape=(len(gen_keys)), dtype=np.float)
        nox_emission_rate = np.empty(shape=(len(gen_keys)), dtype=np.float)


        j = 0
        for key in gen_keys:
            c0 = dict['generators'][key]['c0']
            c1 = dict['generators'][key]['c1']
            c2 = dict['generators'][key]['c2']
            co2_emission_rate[j] = 0
            sox_emission_rate[j] = 0
            nox_emission_rate[j] = 0

            if (dict['generators'][key]['genfuel']) == 'gas':
                if ((dict['generators'][key]['gentype'] == 'combinedcycle') | (
                    dict['generators'][key]['gentype'] == 'combinedcycle')):
                    co2_emission_rate[j] = \
                    gen_emission_rate[dict['generators'][key]['genfuel'] + "_" + dict['generators'][key]['gentype']][0]
                    sox_emission_rate[j] = \
                        gen_emission_rate[
                            dict['generators'][key]['genfuel'] + "_" + dict['generators'][key]['gentype']][1]
                    nox_emission_rate[j] = \
                        gen_emission_rate[
                            dict['generators'][key]['genfuel'] + "_" + dict['generators'][key]['gentype']][2]

            elif (dict['generators'][key]['genfuel']) == 'coal':
                co2_emission_rate[j] = gen_emission_rate[dict['generators'][key]['genfuel']][0]
                sox_emission_rate[j] = gen_emission_rate[dict['generators'][key]['genfuel']][1]
                nox_emission_rate[j] = gen_emission_rate[dict['generators'][key]['genfuel']][2]
            else:
                print('genfuel type', dict['generators'][key]['genfuel'], ' has zero emission or not supported yet!!')

            #print('gen_id, c0, c1, c2:', key, c0, c1, c2)
            i = 0
            for t in times:
                pgen = data_g[j, i, PGEN_IDX]
                gen_cost[j, i] = (c2 * pgen * pgen + c1 * pgen + c0)* time_interval_hours
                gen_revenue[j,i]=data_b[0, i, LMP_P_IDX]*1000.0* pgen * time_interval_hours  # LMP in $/kWh
                #print("gen:", key, co2_emission_rate[j], pgen, time_interval_hours)
                gen_emission_co2[j, i] = co2_emission_rate[j] * pgen * time_interval_hours
                gen_emission_sox[j, i] = sox_emission_rate[j] * pgen * time_interval_hours
                gen_emission_nox[j, i] = nox_emission_rate[j] * pgen * time_interval_hours
                # TODO
                # gen_payment[j,i]=pgen*LMP

                i = i + 1
            j = j + 1

        # print ( "\nco2 rate:",co2_emission_rate)
        # print ( "\nsox rate:",sox_emission_rate)

        self.gen_metrics = xr.Dataset({'PGEN': (['busNum', 'time'], data_g[:, :, PGEN_IDX]),
                                     'QGEN': (['busNum', 'time'], data_g[:, :, QGEN_IDX]),
                                     'COST': (['busNum', 'time'], gen_cost),
                                     'REVENUE': (['busNum', 'time'],gen_revenue),
                                     'EMISSION_CO2': (['busNum', 'time'], gen_emission_co2),
                                     'EMISSION_SOX': (['busNum', 'time'], gen_emission_sox),
                                     'EMISSION_NOX': (['busNum', 'time'], gen_emission_nox),
                                     },
                                    coords={'busNum': list(map(int, gen_keys)),
                                            'time': hrs})  # times or hrs

        ## add the generator dictionary to the attributes of the dataset
        self.gen_metrics.attrs["gen_info_dict"] = dict['generators']
        ##TODO add ancillary generator unit info

    def loadBusMetricsFromNetCMFFile(self, bus_metrics_netCMF):
        self.bus_metrics = xr.open_dataset(bus_metrics_netCMF)


    def loadGenMetricsFromNetCMFFile(self, gen_metrics_netCMF):
        self.gen_metrics = xr.open_dataset(gen_metrics_netCMF)

    def saveBusMetricsToNetCMFFile(self,new_bus_metrics_netCMF):

        self.bus_metrics.to_netcdf(new_bus_metrics_netCMF)
        print('Bus related base metrics are save to : ',new_bus_metrics_netCMF)

    def saveGenMetricsToNetCMFFile(self, new_gen_metrics_netCMF):
        self.gen_metrics.to_netcdf(new_gen_metrics_netCMF)
        print('Gen related base metrics are save to : ', new_gen_metrics_netCMF)

    def get_bus_metrics(self):
        return self.bus_metrics



    def get_bus_metrics_at_time(self, a_time_instance):
        return self.bus_metrics.sel(time = a_time_instance)

    def get_bus_metrics_for_period(self, start_time, end_time):
        return self.bus_metrics.where(np.logical_and(self.bus_metrics.time >= start_time, self.bus_metrics.time < end_time), drop = True)

    def get_bus_metrics_at_bus(self, bus_num):
        return self.bus_metrics.sel(busNum = bus_num)

    def get_allBus_LMP(self):
        return self.bus_metrics.LMP_P

    def get_bus_LMP(self, bus_num):
        return self.bus_metrics.sel(busNum = bus_num).LMP_P

    #
    #-------------------------------- The following are generator related metrics---------------------------------------
    #
    def get_gen_metrics(self):
        return self.gen_metrics
    def get_gen_metrics_for_period(self, start_time, end_time):
        return self.gen_metrics.where(np.logical_and(self.gen_metrics.time >= start_time, self.gen_metrics.time < end_time), drop = True)

    def get_eachGen_cost(self):
        return self.gen_metrics.COST.sum(dim = 'time')

    def get_allGen_cost(self):
        return self.gen_metrics.COST.sum()

    def get_eachGen_revenue(self):
        return self.gen_metrics.REVENUE.sum(dim = 'time')

    def get_allGen_revenue(self):
        return self.gen_metrics.REVENUE.sum()

    def get_eachGen_CO2_emissions(self):
        return self.gen_metrics.EMISSION_CO2.sum(dim='time')

    def get_allGen_CO2_emissions(self):
        return self.gen_metrics.EMISSION_CO2.sum().values

    def get_eachGen_SOx_emissions(self):
        return self.gen_metrics.EMISSION_SOX.sum().sum(dim='time')

    def get_allGen_SOx_emissions(self):
        return self.gen_metrics.EMISSION_SOX.sum().values

    def get_eachGen_NOx_emissions(self):
        return self.gen_metrics.EMISSION_NOX.sum().sum(dim='time')

    def get_allGen_NOx_emissions(self):
        return self.gen_metrics.EMISSION_NOX.sum().values

    def get_cost_at_gen(self, busNum):
        return self.gen_metrics.sel(busNum = busNum).GENCOST.sum(dim='time')


# Copyright (C) 2017-2020 Battelle Memorial Institute
# file: api.py
"""Functions intended for public access.

Example:
    To start PYPOWER for connection to FNCS::

        import tesp_support.api as tesp
        tesp.pypower_loop('te30_pp.json','TE_Challenge')

Public Functions:
    :substation_loop: Supervises FNCS messages and time steps for a GridLAB-D substation with one double-auction market and many HVAC controllers.  
    :pypower_loop: Supervises the FNCS messages, time steps, optimal power flow and power flow for PYPOWER.  
    :precool_loop: Supervises FNCS messages and time steps for a GridLAB-D substation with many price-taking and pre-cooling HVAC controllers.  
    :prep_substation: Writes agent metadata and FNCS subscriptions used in substation_loop.  
    :prep_precool: Writes agent metadata and FNCS subscriptions used in precool_loop.  
    :glm_dict: Writes the JSON metadata from a GridLAB-D file.
    :populate_feeder: Replaces the ZIP loads with houses in a GridLAB-D file.
    :make_tesp_case: Makes a subdirectory with all files needed to run a TESP simulation, as directed by a JSON file.
    :make_monte_carlo_cases: Makes a subdirectory with all files needed to run a Monte Carlo TESP simulation, as directed by a JSON file.
    :convert_tmy2_to_epw: Command line utility that converts TMY2 weather files to the EPW format for EnergyPlus.
    :weathercsv: Converts TMY3 weather data to CSV format.
    :startWeatherAgent: Publishes FNCS messages and forecasts from a weather CSV file.
    :make_ems: Creates the energy management system (EMS) for FNCS/HELICS to interface with EnergyPlus.

"""

from __future__ import absolute_import

from .feederGenerator import populate_feeder
from .fncsPYPOWER import pypower_loop 
from .glm_dict import glm_dict
from .precool import precool_loop
from .prep_precool import prep_precool
from .prep_substation import prep_substation
from .tesp_case import make_tesp_case
from .tesp_case import make_monte_carlo_cases
from .tesp_case import add_tesp_feeder
from .TMY2EPW import convert_tmy2_to_epw
from .TMY3toCSV import weathercsv 
from .substation import substation_loop
from .weatherAgent import startWeatherAgent

from .case_merge import merge_glm
from .case_merge import merge_glm_dict
from .case_merge import merge_agent_dict
from .case_merge import merge_substation_yaml
from .case_merge import merge_fncs_config
from .case_merge import merge_gld_msg
from .case_merge import merge_substation_msg
 
from .fncsPYPOWER import load_json_case
from .fncsPYPOWER import summarize_opf

from .make_ems import make_ems
from .make_ems import merge_idf

#from .process_agents import process_agents
#from .process_eplus import process_eplus
#from .process_gld import process_gld
#from .process_houses import process_houses
#from .process_inv import process_inv
#from .process_pypower import process_pypower
#from .process_voltages import process_voltages

#from .matpower.matpower_dict import matpower_dict
#from .matpower.process_matpower import process_matpower

#from .sgip1.compare_auction import compare_auction
#from .sgip1.compare_csv import compare_csv
#from .sgip1.compare_hvac import compare_hvac
#from .sgip1.compare_prices import compare_prices
#from .sgip1.compare_pypower import compare_pypower

#from .valuation.TransmissionMetricsProcessor import TransmissionMetricsProcessor





# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: process_voltages.py
"""Functions to plot all billing meter voltages from GridLAB-D

Public Functions:
    :process_voltages: Reads the data and metadata, then makes the plot.  

"""
import json;
import sys;
import numpy as np;
try:
  import matplotlib as mpl;
  import matplotlib.pyplot as plt;
except:
  pass

def process_voltages(nameroot, dictname = ''):
  """ Plots the min and max line-neutral voltages for every billing meter

  This function reads *substation_nameroot_metrics.json* and 
  *billing_meter_nameroot_metrics.json* for the voltage data, and 
  *nameroot_glm_dict.json* for the meter names.  
  These must all exist in the current working directory.
  One graph is generated with 2 subplots:
  
  1. The Min line-to-neutral voltage at each billing meter  
  2. The Max line-to-neutral voltage at each billing meter  

  Args:
    nameroot (str): name of the TESP case, not necessarily the same as the GLM case, without the extension
    dictname (str): metafile name (with json extension) for a different GLM dictionary, if it's not *nameroot_glm_dict.json*. Defaults to empty.
  """
  # first, read and print a dictionary of all the monitored GridLAB-D objects
  if len (dictname) > 0:
      lp = open (dictname).read()
  else:
      lp = open (nameroot + "_glm_dict.json").read()
  dict = json.loads(lp)
  mtr_keys = list(dict['billingmeters'].keys())
  mtr_keys.sort()
# print("\nBilling Meter Dictionary:")
# for key in mtr_keys:
#   row = dict['billingmeters'][key]
# # print (key, "on phase", row['phases'], "of", row['feeder_id'], "with", row['children'])

  # make a sorted list of the sample times in hours
  lp_m = open ("billing_meter_" + nameroot + "_metrics.json").read()
  lst_m = json.loads(lp_m)
  lst_m.pop('StartTime')
  meta_m = lst_m.pop('Metadata')
  times = list(map(int,list(lst_m.keys())))
  times.sort()
  print ("There are", len (times), "sample times at", times[1] - times[0], "second intervals")
  hrs = np.array(times, dtype=np.float)
  denom = 3600.0
  hrs /= denom

  #print("\nBilling Meter Metadata for", len(lst_m['3600']), "objects")
  for key, val in meta_m.items():
  # print (key, val['index'], val['units'])
    if key == 'voltage_max':
      MTR_VOLT_MAX_IDX = val['index']
      MTR_VOLT_MAX_UNITS = val['units']
    elif key == 'voltage_min':
      MTR_VOLT_MIN_IDX = val['index']
      MTR_VOLT_MIN_UNITS = val['units']
    elif key == 'voltage_avg':
      MTR_VOLT_AVG_IDX = val['index']
      MTR_VOLT_AVG_UNITS = val['units']
    elif key == 'voltage12_max':
      MTR_VOLT12_MAX_IDX = val['index']
      MTR_VOLT12_MAX_UNITS = val['units']
    elif key == 'voltage12_min':
      MTR_VOLT12_MIN_IDX = val['index']
      MTR_VOLT12_MIN_UNITS = val['units']
    elif key == 'voltage_unbalance_max':
      MTR_VOLTUNB_MAX_IDX = val['index']
      MTR_VOLTUNB_MAX_UNITS = val['units']

  time_key = str(times[0])
  data_m = np.empty(shape=(len(mtr_keys), len(times), len(lst_m[time_key][mtr_keys[0]])), dtype=np.float)
  print ("\nConstructed", data_m.shape, "NumPy array for Meters")
  j = 0
  for key in mtr_keys:
    i = 0
    for t in times:
      ary = lst_m[str(t)][mtr_keys[j]]
      data_m[j, i,:] = ary
      i = i + 1
    j = j + 1

  # normalize the meter voltages to 100 percent
  j = 0
  for key in mtr_keys:
    vln = dict['billingmeters'][key]['vln'] / 100.0
    vll = dict['billingmeters'][key]['vll'] / 100.0
    data_m[j,:,MTR_VOLT_MIN_IDX] /= vln
    data_m[j,:,MTR_VOLT_MAX_IDX] /= vln
    data_m[j,:,MTR_VOLT_AVG_IDX] /= vln
    data_m[j,:,MTR_VOLT12_MIN_IDX] /= vll
    data_m[j,:,MTR_VOLT12_MAX_IDX] /= vll
    j = j + 1

  # display a plot
  fig, ax = plt.subplots(2, 1, sharex = 'col')
  i = 0
  for key in mtr_keys:
    ax[0].plot(hrs, data_m[i,:,MTR_VOLT_MIN_IDX], color="blue")
    ax[1].plot(hrs, data_m[i,:,MTR_VOLT_MAX_IDX], color="red")
    i = i + 1
  ax[0].set_ylabel("Min Voltage [%]")
  ax[1].set_ylabel("Max Voltage [%]")
  ax[1].set_xlabel("Hours")
  ax[0].set_title ("Voltage at all Meters")

  plt.show()





#   Copyright (C) 2017-2019 Battelle Memorial Institute
# file: process_eplus.py
"""Functions to plot data from the EnergyPlus agent

Public Functions:
        :process_eplus: Reads the data and metadata, then makes the plots.  

"""
import json;
import sys;
import numpy as np;
try:
  import matplotlib as mpl;
  import matplotlib.pyplot as plt;
except:
  pass

def process_eplus(nameroot, title=None, pngfile=None):
    """ Plots the min and max line-neutral voltages for every billing meter

    This function reads *eplus_nameroot_metrics.json* for both metadata and data. 
    This must exist in the current working directory.  
    One graph is generated with 3 subplots:

    1. Cooling system setpoint, actual temperature and the difference between them.
    2. Heating system setpoint, actual temperature and the difference between them.
    3. Price that the building controller responded to.

    Args:
        nameroot (str): name of the TESP case, not necessarily the same as the EnergyPlus case, without the extension
    """
    # read the JSON file
    try:
      lp = open ("eplus_" + nameroot + "_metrics.json").read()
      lst = json.loads(lp)
      print ("Metrics data starting", lst['StartTime'])
    except:
      print ('eplus metrics file could not be read')
      return

    # make a sorted list of the times
    lst.pop('StartTime')
    meta = lst.pop('Metadata')
    times = list(map(int,list(lst.keys())))
    times.sort()
    print ("There are", len (times), "sample times at", times[1] - times[0], "seconds")

    # parse the metadata for things of specific interest
    for key, val in meta.items():
        if key == 'electric_demand_power_avg':
            ELECTRIC_DEMAND_IDX = val['index']
            ELECTRIC_DEMAND_UNITS = val['units']
        elif key == 'hvac_demand_power_avg':
            HVAC_DEMAND_IDX = val['index']
            HVAC_DEMAND_UNITS = val['units']
        elif key == 'occupants_total_avg':
            OCCUPANTS_IDX = val['index']
            OCCUPANTS_UNITS = val['units']
        elif key == 'kwhr_price_avg':
            PRICE_IDX = val['index']
            PRICE_UNITS = val['units']
        elif key == 'ashrae_uncomfortable_hours_avg':
            ASHRAE_HOURS_IDX = val['index']
            ASHRAE_HOURS_UNITS = val['units']
        elif key == 'cooling_schedule_temperature_avg':
            COOLING_SCHEDULE_IDX = val['index']
            COOLING_SCHEDULE_UNITS = val['units']
        elif key == 'cooling_setpoint_temperature_avg':
            COOLING_SETPOINT_IDX = val['index']
            COOLING_SETPOINT_UNITS = val['units']
        elif key == 'cooling_current_temperature_avg':
            COOLING_TEMPERATURE_IDX = val['index']
            COOLING_TEMPERATURE_UNITS = val['units']
        elif key == 'cooling_setpoint_delta_avg':
            COOLING_DELTA_IDX = val['index']
            COOLING_DELTA_UNITS = val['units']
        elif key == 'cooling_controlled_load_avg':
            COOLING_POWER_IDX = val['index']
            COOLING_POWER_UNITS = val['units']
        elif key == 'cooling_power_state_avg':
            COOLING_STATE_IDX = val['index']
            COOLING_STATE_UNITS = val['units']
        elif key == 'heating_schedule_temperature_avg':
            HEATING_SCHEDULE_IDX = val['index']
            HEATING_SCHEDULE_UNITS = val['units']
        elif key == 'heating_setpoint_temperature_avg':
            HEATING_SETPOINT_IDX = val['index']
            HEATING_SETPOINT_UNITS = val['units']
        elif key == 'heating_current_temperature_avg':
            HEATING_TEMPERATURE_IDX = val['index']
            HEATING_TEMPERATURE_UNITS = val['units']
        elif key == 'heating_setpoint_delta_avg':
            HEATING_DELTA_IDX = val['index']
            HEATING_DELTA_UNITS = val['units']
        elif key == 'heating_controlled_load_avg':
            HEATING_POWER_IDX = val['index']
            HEATING_POWER_UNITS = val['units']
        elif key == 'heating_power_state_avg':
            HEATING_STATE_IDX = val['index']
            HEATING_STATE_UNITS = val['units']
        elif key == 'outdoor_air_avg':
            OUTDOOR_AIR_IDX = val['index']
            OUTDOOR_AIR_UNITS = val['units']
        elif key == 'indoor_air_avg':
            INDOOR_AIR_IDX = val['index']
            INDOOR_AIR_UNITS = val['units']
        elif key == 'heating_volume_avg':
            HEATING_VOLUME_IDX = val['index']
            HEATING_VOLUME_UNITS = val['units']
        elif key == 'cooling_volume_avg':
            COOLING_VOLUME_IDX = val['index']
            COOLING_VOLUME_UNITS = val['units']


    # make sure we found the metric indices of interest
    building = list(lst['3600'].keys())[0]
    ary = lst['3600'][building]
    print ('There are', len(ary), 'metrics for', building)
    print ('1st hour price =', ary[PRICE_IDX], PRICE_UNITS)

    # create a NumPy array of all metrics for the first building, 8760*39 doubles
    # we also want a NumPy array of times in hours
    data = np.empty(shape=(len(times),len(ary)), dtype=np.float)
    print ('Constructed', data.shape, 'NumPy array')
    i = 0
    for t in times:
        ary = lst[str(t)][building]
        data[i,:] = ary
        i = i + 1
    hrs = np.array(times, dtype=np.float)
    denom = 3600.0
    hrs /= denom

    # display some averages
    print ('Average price  = {:.5f}'.format (data[:,PRICE_IDX].mean()), PRICE_UNITS)
    print ('Average demand = {:.2f}'.format (data[:,ELECTRIC_DEMAND_IDX].mean()), ELECTRIC_DEMAND_UNITS)
    print ('Average HVAC   = {:.2f}'.format (data[:,HVAC_DEMAND_IDX].mean()), HVAC_DEMAND_UNITS)
    print ('Average uncomf = {:.5f}'.format (data[:,ASHRAE_HOURS_IDX].mean()), ASHRAE_HOURS_UNITS)
    print ('Average people = {:.2f}'.format (data[:,OCCUPANTS_IDX].mean()), OCCUPANTS_UNITS)

    print ('Average cooling power = {:9.2f}'.format (data[:,COOLING_POWER_IDX].mean()), COOLING_POWER_UNITS)
    print ('Average cooling temp  = {:9.2f}'.format (data[:,COOLING_TEMPERATURE_IDX].mean()), COOLING_TEMPERATURE_UNITS)
    print ('Average cooling sched = {:9.2f}'.format (data[:,COOLING_SCHEDULE_IDX].mean()), COOLING_SCHEDULE_UNITS)
    print ('Average cooling delta = {:9.2f}'.format (data[:,COOLING_DELTA_IDX].mean()), COOLING_DELTA_UNITS)
    print ('Average cooling setpt = {:9.2f}'.format (data[:,COOLING_SETPOINT_IDX].mean()), COOLING_SETPOINT_UNITS)
    print ('Average cooling vol   = {:9.2f}'.format (data[:,COOLING_VOLUME_IDX].mean()), COOLING_VOLUME_UNITS)

    print ('Average heating power = {:9.2f}'.format (data[:,HEATING_POWER_IDX].mean()), HEATING_POWER_UNITS)
    print ('Average heating temp  = {:9.2f}'.format (data[:,HEATING_TEMPERATURE_IDX].mean()), HEATING_TEMPERATURE_UNITS)
    print ('Average heating sched = {:9.2f}'.format (data[:,HEATING_SCHEDULE_IDX].mean()), HEATING_SCHEDULE_UNITS)
    print ('Average heating delta = {:9.2f}'.format (data[:,HEATING_DELTA_IDX].mean()), HEATING_DELTA_UNITS)
    print ('Average heating setpt = {:9.2f}'.format (data[:,HEATING_SETPOINT_IDX].mean()), HEATING_SETPOINT_UNITS)
    print ('Average heating vol   = {:9.2f}'.format (data[:,HEATING_VOLUME_IDX].mean()), HEATING_VOLUME_UNITS)

    print ('Average outdoor air   = {:9.2f}'.format (data[:,OUTDOOR_AIR_IDX].mean()), OUTDOOR_AIR_UNITS)
    print ('Average indoor air    = {:9.2f}'.format (data[:,INDOOR_AIR_IDX].mean()), INDOOR_AIR_UNITS)

    # display a plot
    width = 12.0
    height = 8.0
    fig, ax = plt.subplots(3,3, sharex = 'col', figsize=(width,height), constrained_layout=True)
    if title is not None:
      fig.suptitle (title)

    ax[0,0].plot(hrs, data[:,COOLING_TEMPERATURE_IDX], color='blue', label='Actual')
    ax[0,0].plot(hrs, data[:,COOLING_SETPOINT_IDX], color='red', label='Setpoint')
    ax[0,0].plot(hrs, data[:,COOLING_SCHEDULE_IDX], color='green', label='Schedule')
#    ax[0,0].plot(hrs, data[:,INDOOR_AIR_IDX], color='magenta', label='Indoor')
    ax[0,0].set_ylabel(COOLING_TEMPERATURE_UNITS)
    ax[0,0].set_title ('Volume Average Cooling')
    ax[0,0].legend(loc='best')

    ax[1,0].plot(hrs, data[:,HEATING_TEMPERATURE_IDX], color='blue', label='Actual')
    ax[1,0].plot(hrs, data[:,HEATING_SETPOINT_IDX], color='red', label='Setpoint')
    ax[1,0].plot(hrs, data[:,HEATING_SCHEDULE_IDX], color='green', label='Schedule')
#    ax[1,0].plot(hrs, data[:,INDOOR_AIR_IDX], color='magenta', label='Indoor')
    ax[1,0].set_ylabel(HEATING_TEMPERATURE_UNITS)
    ax[1,0].set_title ('Volume Average Heating')
    ax[1,0].legend(loc='best')

    ax[2,0].plot(hrs, data[:,OUTDOOR_AIR_IDX], color='blue', label='Outdoor')
    ax[2,0].plot(hrs, data[:,INDOOR_AIR_IDX], color='red', label='Indoor')
    ax[2,0].set_ylabel(OUTDOOR_AIR_UNITS)
    ax[2,0].set_title ('Average Temperatures')
    ax[2,0].legend(loc='best')

    ax[0,1].plot(hrs, data[:,PRICE_IDX], color='blue', label='Actual')
    ax[0,1].set_ylabel(PRICE_UNITS)
    ax[0,1].set_title ('Real-time Price')

    ax[1,1].plot(hrs, 0.001 * data[:,ELECTRIC_DEMAND_IDX], color='blue', label='Total')
    ax[1,1].plot(hrs, 0.001 * data[:,HVAC_DEMAND_IDX], color='red', label='HVAC')
    ax[1,1].set_ylabel('kW')
    ax[1,1].set_title ('Building Electrical Demand')
    ax[1,1].legend(loc='best')

#    ax[2,1].plot(hrs, data[:,COOLING_STATE_IDX], color='blue', label='Cooling')
#    ax[2,1].plot(hrs, data[:,HEATING_STATE_IDX], color='red', label='Heating')
#    ax[2,1].set_ylabel(HEATING_STATE_UNITS)
    ax[2,1].plot(hrs, 0.001 * data[:,COOLING_POWER_IDX], color='blue', label='Cooling')
    ax[2,1].plot(hrs, 0.001 * data[:,HEATING_POWER_IDX], color='red', label='Heating')
    ax[2,1].plot(hrs, 0.001 * data[:,HVAC_DEMAND_IDX], color='green', label='HVAC')
    ax[2,1].set_ylabel('kW')
    ax[2,1].set_title ('DX/Electrical Coil Demand')
    ax[2,1].legend(loc='best')

    ax[0,2].plot(hrs, data[:,OCCUPANTS_IDX], color='blue')
    ax[0,2].set_ylabel(OCCUPANTS_UNITS)
    ax[0,2].set_title ('Occupants')

    ax[1,2].plot(hrs, data[:,ASHRAE_HOURS_IDX], color='blue')
    ax[1,2].set_ylabel(ASHRAE_HOURS_UNITS)
    ax[1,2].set_title ('Uncomfortable Hours')

    ax[2,2].plot(hrs, 0.001 * data[:,HEATING_VOLUME_IDX] + 0.001 * data[:,COOLING_VOLUME_IDX], color='magenta', label='Total')
    ax[2,2].plot(hrs, 0.001 * data[:,COOLING_VOLUME_IDX], color='blue', label='Cooling')
    ax[2,2].plot(hrs, 0.001 * data[:,HEATING_VOLUME_IDX], color='red', label='Heating')
    ax[2,2].set_ylabel('thousand m^3')
    ax[2,2].set_title ('Sensible Zone Volumes')
    ax[2,2].legend(loc='best')

    ax[2,0].set_xlabel('Hours')
    ax[2,1].set_xlabel('Hours')
    ax[2,2].set_xlabel('Hours')

    if pngfile is not None:
      plt.savefig(pngfile)
    else:
      plt.show()




import csv;
import sys;
import numpy as np;
import matplotlib as mpl;
import matplotlib.pyplot as plt;

# here are the columns, after ignore 'True/False' with usecols
plotvars = [[0, 't[s]', 'seconds'],
						[1, 'Pload', 'MW'],
						[2, 'P7 (csv)', 'MW'],
						[3, 'GLD Unresp', 'MW'],
						[4, 'P7 (opf)', 'MW'],
						[5, 'Resp (opf)', 'MW'],
						[6, 'GLD Pub', 'MW'],
						[7, 'P7 Min', 'MW'],
						[8, 'V7', 'pu'],
						[9, 'LMP_P7', '$/MWHR'],
						[10, 'LMP_Q7', '$/MWHR'],
						[11, 'Pgen1', 'MW'],
						[12, 'Pgen2', 'MW'],
						[13, 'Pgen3', 'MW'],
						[14, 'Pgen4', 'MW'],
						[15, 'Pdisp', 'MW'],
						[16, 'gencost2', '$'],
						[17, 'gencost1', '$'],
						[18, 'gencost0', '$']]

def UnitPrice (p, c2, c1, c0):
	lmp = p * 2.0 * c2 + c1
	return lmp

def MakePlotData(root):
	d1 = np.loadtxt(root + '.csv', skiprows=1, delimiter=',',
									usecols=[0,2,3,4,5,6,7,9,10,11,12,13,14,15,16,17,18,19,20])
	hrs = d1[:,0] / 3600.0
	lmp = d1[:,9]
	p1 = d1[:,11]
	p2 = d1[:,12]
	p3 = d1[:,13]
	p4 = d1[:,14]
	pdisp = d1[:,15]
	gc2 = d1[:,16]
	gc1 = d1[:,17]

	lmp1 = UnitPrice (p1, 0.030,  1.3,  180.0)
	lmp2 = UnitPrice (p2, 0.085,  3.0,  100.0)
	lmp3 = UnitPrice (p3, 0.122,  5.0,  135.0)
	lmp4 = UnitPrice (p4, 3.000, 50.0, 1500.0)
	lmp5 = 2.0 * pdisp * gc2 + gc1

	return hrs, lmp, lmp1, lmp2, lmp3, lmp4, lmp5

def compare_prices (rootname):
	# display a plot
	tmin = 0.0
	tmax = 48.0
	xticks = [0,6,12,18,24,30,36,42,48]

	fig, ax = plt.subplots(1, 1, sharex = 'col')

	hrs, lmp, lmp1, lmp2, lmp3, lmp4, lmp5 = MakePlotData (rootname)
	#ax.plot(hrs, lmp1, color='blue', label='LMP1')
	#ax.plot(hrs, lmp2, color='green', label='LMP2')
	#ax.plot(hrs, lmp3, color='magenta', label='LMP3')
	#ax.plot(hrs, lmp4, color='cyan', label='LMP4')
	ax.plot(hrs, lmp5, color='orange', label='LMP5')
	ax.plot(hrs, lmp, color='red', label='LMP')

	ax.set_title ('Price Comparison for ' + root)
	ax.set_ylabel ('$/MWHR')
	ax.set_xlabel ('Hours')
	ax.grid()
	ax.legend()
	ax.set_xlim(tmin,tmax)
	ax.set_xticks(xticks)

	plt.show()





# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: simple_auction.py
"""Double-auction mechanism for the 5-minute markets in te30 and sgip1 examples

The substation_loop module manages one instance of this class per GridLAB-D substation.

Todo:
    * Initialize and update price history statistics
    * Allow for adjustment of clearing_scalar
    * Handle negative price bids from HVAC agents, currently they are discarded
    * Distribute marginal quantities and fractions; these are not currently applied to HVACs

"""
import tesp_support.helpers as helpers

# Class definition
class simple_auction:
    """This class implements a simplified version of the double-auction market embedded in GridLAB-D.

    References:
        `Market Module Overview - Auction <http://gridlab-d.shoutwiki.com/wiki/Market_Auction>`_

    Args:
        dict (dict): a row from the agent configuration JSON file
        key (str): the name of this agent, which is the market key from the agent configuration JSON file

    Attributes:
        name (str): the name of this auction, also the market key from the configuration JSON file
        std_dev (float): the historical standard deviation of the price, in $/kwh, from dict
        mean (float): the historical mean price in $/kwh, from dict
        period (float): the market clearing period in seconds, from dict
        pricecap (float): the maximum allowed market clearing price, in $/kwh, from dict
        max_capacity_reference_bid_quantity (float):
        statistic_mode (int): always 1, not used, from dict
        stat_mode (str): always ST_CURR, not used, from dict
        stat_interval (str): always 86400 seconds, for one day, not used, from dict
        stat_type (str): always mean and standard deviation, not used, from dict
        stat_value (str): always zero, not used, from dict
        curve_buyer (curve): data structure to accumulate buyer bids
        curve_seller (curve): data structure to accumulate seller bids
        refload (float): the latest substation load from GridLAB-D
        lmp (float): the latest locational marginal price from the bulk system market
        unresp (float): unresponsive load, i.e., total substation load less the bidding, running HVACs
        agg_unresp (float): aggregated unresponsive load, i.e., total substation load less the bidding, running HVACs
        agg_resp_max (float): total load of the bidding HVACs
        agg_deg (int): degree of the aggregate bid curve polynomial, should be 0 (zero or one bids), 1 (2 bids) or 2 (more bids)
        agg_c2 (float): second-order coefficient of the aggregate bid curve
        agg_c1 (float): first-order coefficient of the aggregate bid curve
        clearing_type (helpers.ClearingType): describes the solution type or boundary case for the latest market clearing
        clearing_quantity (float): quantity at the last market clearing
        clearing_price (float): price at the last market clearing
        marginal_quantity (float): quantity of a partially accepted bid
        marginal_frac (float): fraction of the bid quantity accepted from a marginal buyer or seller 
        clearing_scalar (float): used for interpolation at boundary cases, always 0.5
    """
    # ====================Define instance variables ===================================
    def __init__(self,dict,key):
        self.name = key
        self.std_dev = float(dict['init_stdev'])
        self.mean = float(dict['init_price'])
        self.period = float(dict['period'])
        self.pricecap = float(dict['pricecap'])
        self.max_capacity_reference_bid_quantity = float(dict['max_capacity_reference_bid_quantity'])
        self.statistic_mode = int(dict['statistic_mode'])
        self.stat_mode = dict['stat_mode']
        self.stat_interval = dict['stat_interval']
        self.stat_type = dict['stat_type']
        self.stat_value = dict['stat_value']

        # updated in collect_agent_bids, used in clear_market
        self.curve_buyer = None
        self.curve_seller = None

        self.refload = 0.0
        self.lmp = self.mean
        self.unresp = 0.0
        self.agg_unresp = 0.0
        self.agg_resp_max = 0.0
        self.agg_deg = 0
        self.agg_c2 = 0.0
        self.agg_c1 = 0.0
        self.clearing_type = helpers.ClearingType.NULL
        self.clearing_quantity = 0
        self.clearing_price = self.mean
        self.marginal_quantity = 0.0
        self.marginal_frac = 0.0

        self.clearing_scalar = 0.5

        self.consumerSurplus = 0.0
        self.averageConsumerSurplus = 0.0
        self.supplierSurplus = 0.0
        self.unrespSupplierSurplus = 0.0

        self.bid_offset = 1e-4 # for numerical checks

    def set_refload (self, kw):
        """Sets the refload attribute

        Args:
            kw (float): GridLAB-D substation load in kw
        """
        self.refload = kw

    def set_lmp (self, lmp):
        """Sets the lmp attribute

        Args:
            lmp (float): locational marginal price from the bulk system market
        """
        self.lmp = lmp
                
    def initAuction (self):
        """Sets the clearing_price and lmp to the mean price
        """
        self.clearing_price = self.lmp = self.mean

    def update_statistics (self):
        """Update price history statistics - not implemented
        """
        sample_need = 0

    def clear_bids (self):
        """Re-initializes curve_buyer and curve_seller, sets the unresponsive load estimate to the total substation load.
        """
        self.curve_buyer = helpers.curve ()
        self.curve_seller = helpers.curve ()
        self.unresp = self.refload
    
    def supplier_bid (self, bid):
        """Gather supplier bids into curve_seller

        Use this to enter curves in step-wise blocks.

        Args:
            bid ([float, float]): price in $/kwh, quantity in kW
        """
        price = bid[0]
        quantity = bid[1]
        self.curve_seller.add_to_curve (price, quantity, False)

    def collect_bid (self, bid):
        """Gather HVAC bids into curve_buyer

        Also adjusts the unresponsive load estimate, by subtracting the HVAC power
        if the HVAC is on.

        Args:
            bid ([float, float, Boolean]): price in $/kwh, quantity in kW and the HVAC on state
        """
        price = bid[0]
        quantity = bid[1]
        is_on = bid[2]
        if is_on:
            self.unresp -= quantity
        if price > 0.0:
            self.curve_buyer.add_to_curve (price, quantity, is_on)

    def aggregate_bids (self):
        """Aggregates the unresponsive load and responsive load bids for submission to the bulk system market
        """
        if self.unresp > 0:
            self.curve_buyer.add_to_curve (self.pricecap, self.unresp, True)
        else:
            print ('$$ flag,Unresp,BuyCount,BuyTotal,BuyOn,BuyOff', flush=True)
            print ('$$ unresp < 0', 
                   '{:.3f}'.format(self.unresp), 
                   self.curve_buyer.count, 
                   '{:.3f}'.format(self.curve_buyer.total), 
                   '{:.3f}'.format(self.curve_buyer.total_on), 
                   '{:.3f}'.format(self.curve_buyer.total_off), 
                   sep=',', flush=True)
        if self.curve_buyer.count > 0:
            self.curve_buyer.set_curve_order ('descending')
        self.agg_unresp, self.agg_resp_max, self.agg_deg, self.agg_c2, self.agg_c1 = helpers.aggregate_bid (self.curve_buyer)

    def clear_market (self, tnext_clear=0, time_granted=0):
        """Solves for the market clearing price and quantity

        Uses the current contents of curve_seller and curve_buyer.
        Updates clearing_price, clearing_quantity, clearing_type,
        marginal_quantity and marginal_frac.

        Args:
            tnext_clear (int): next clearing time in FNCS seconds, should be <= time_granted, for the log file only
            time_granted (int): the current time in FNCS seconds, for the log file only
        """
        if self.max_capacity_reference_bid_quantity > 0:
            self.curve_seller.add_to_curve (self.lmp, self.max_capacity_reference_bid_quantity, True)
        if self.curve_seller.count > 0:
            self.curve_seller.set_curve_order ('ascending')

        self.unresponsive_sell = self.responsive_sell = self.unresponsive_buy = self.responsive_buy = 0

        if self.curve_buyer.count > 0 and self.curve_seller.count > 0:
            a = self.pricecap
            b = -self.pricecap
            check = 0
            demand_quantity = supply_quantity = 0
            for i in range(self.curve_seller.count):
                if self.curve_seller.price[i] == self.pricecap:
                    self.unresponsive_sell += self.curve_seller.quantity[i]
                else:
                    self.responsive_sell += self.curve_seller.quantity[i]
            for i in range(self.curve_buyer.count):
                if self.curve_buyer.price[i] == self.pricecap:
                    self.unresponsive_buy += self.curve_buyer.quantity[i]
                else:
                    self.responsive_buy += self.curve_buyer.quantity[i]
            # Calculate clearing quantity and price here
            # Define the section number of the buyer and the seller curves respectively as i and j
            i = j = 0
            self.clearing_type = helpers.ClearingType.NULL
            self.clearing_quantity = self.clearing_price = 0
            while i < self.curve_buyer.count and j < self.curve_seller.count and self.curve_buyer.price[i] >= self.curve_seller.price[j]:
                buy_quantity = demand_quantity + self.curve_buyer.quantity[i]
                sell_quantity = supply_quantity + self.curve_seller.quantity[j]
                # If marginal buyer currently:
                if buy_quantity > sell_quantity:
                    self.clearing_quantity = supply_quantity = sell_quantity
                    a = b = self.curve_buyer.price[i]
                    j += 1
                    check = 0
                    self.clearing_type = helpers.ClearingType.BUYER
                # If marginal seller currently:
                elif buy_quantity < sell_quantity:
                    self.clearing_quantity = demand_quantity = buy_quantity
                    a = b = self.curve_seller.price[j]
                    i += 1
                    check = 0
                    self.clearing_type = helpers.ClearingType.SELLER
                # Buy quantity equal sell quantity but price split  
                else:
                    self.clearing_quantity = demand_quantity = supply_quantity = buy_quantity
                    a = self.curve_buyer.price[i]
                    b = self.curve_seller.price[j]
                    i += 1
                    j += 1
                    check = 1
            # End of the curve comparison, and if EXACT, get the clear price
            if a == b:
                self.clearing_price = a 
            # If there was price agreement or quantity disagreement
            if check:
                self.clearing_price = a 
                if supply_quantity == demand_quantity:
                    # At least one side exhausted at same quantity
                    if i == self.curve_buyer.count or j == self.curve_seller.count:
                        if a == b:
                            self.clearing_type = helpers.ClearingType.EXACT
                        else:
                            self.clearing_type = helpers.ClearingType.PRICE
                    # Exhausted buyers, sellers unsatisfied at same price
                    elif i == self.curve_buyer.count and b == self.curve_seller.price[j]:
                        self.clearing_type = helpers.ClearingType.SELLER
                    # Exhausted sellers, buyers unsatisfied at same price
                    elif j == self.curve_seller.count and a == self.curve_buyer.price[i]:
                        self.clearing_type = helpers.ClearingType.BUYER
                    # Both sides satisfied at price, but one side exhausted  
                    else:
                        if a == b:
                            self.clearing_type = helpers.ClearingType.EXACT
                        else:
                            self.clearing_type = helpers.ClearingType.PRICE
                # No side exausted
                else:
                    # Price changed in both directions
                    if a != self.curve_buyer.price[i] and b != self.curve_seller.price[j] and a == b:
                        self.clearing_type = helpers.ClearingType.EXACT
                    # Sell price increased ~ marginal buyer since all sellers satisfied
                    elif a == self.curve_buyer.price[i] and b != self.curve_seller.price[j]:
                        self.clearing_type = helpers.ClearingType.BUYER
                    # Buy price increased ~ marginal seller since all buyers satisfied
                    elif a != self.curve_buyer.price[i] and b == self.curve_seller.price[j]:
                        self.clearing_type = helpers.ClearingType.SELLER
                        self.clearing_price = b # use seller's price, not buyer's price
                    # Possible when a == b, q_buy == q_sell, and either the buyers or sellers are exhausted
                    elif a == self.curve_buyer.price[i] and b == self.curve_seller.price[j]:
                        if i == self.curve_buyer.count and j == self.curve_seller.count:
                            self.clearing_type = helpers.ClearingType.EXACT
                        elif i == self.curve_buyer.count:
                            self.clearing_type = helpers.ClearingType.SELLER
                        elif j == self.curve_seller.count:
                            self.clearing_type = helpers.ClearingType.BUYER
                    else:
                        # Marginal price
                        self.clearing_type = helpers.ClearingType.PRICE
                
                # If ClearingType.PRICE, calculate the clearing price here
                dHigh = dLow = 0
                if self.clearing_type == helpers.ClearingType.PRICE:
                    avg = (a+b)/2.0
                    # Calculating clearing price limits:   
                    dHigh = a if i == self.curve_buyer.count else self.curve_buyer.price[i]
                    dLow = b if j == self.curve_seller.count else self.curve_seller.price[j]
                    # Needs to be just off such that it does not trigger any other bids
                    if a == self.pricecap and b != -self.pricecap:
                        if self.curve_buyer.price[i] > b:
                            self.clearing_price = self.curve_buyer.price[i] + self.bid_offset
                        else:
                            self.clearing_price = b 
                    elif a != self.pricecap and b == -self.pricecap:
                        if self.curve_seller.price[j] < a:
                            self.clearing_price = self.curve_seller.price[j] - self.bid_offset
                        else:
                            self.clearing_price = a 
                    elif a == self.pricecap and b == -self.pricecap:
                        if i == self.curve_buyer.count and j == self.curve_seller.count:
                            self.clearing_price = 0 # no additional bids on either side
                        elif j == self.curve_seller.count: # buyers left
                            self.clearing_price = self.curve_buyer.price[i] + self.bid_offset
                        elif i == self.curve_buyer.count: # sellers left
                            self.clearing_price = self.curve_seller.price[j] - self.bid_offset
                        else: # additional bids on both sides, just no clearing
                            self.clearing_price = (dHigh + dLow)/2
                    else:
                        if i != self.curve_buyer.count and self.curve_buyer.price[i] == a:
                            self.clearing_price = a 
                        elif j != self.curve_seller.count and self.curve_seller.price[j] == b:
                            self.clearing_price = b 
                        elif i != self.curve_buyer.count and avg < self.curve_buyer.price[i]:
                            self.clearing_price = dHigh + self.bid_offset
                        elif j != self.curve_seller.count and avg > self.curve_seller.price[j]:
                            self.clearing_price = dLow - self.bid_offset
                        else:
                            self.clearing_price = avg 
                                
            # Check for zero demand but non-zero first unit sell price
            if self.clearing_quantity == 0:
                self.clearing_type = helpers.ClearingType.NULL
                if self.curve_seller.count > 0 and self.curve_buyer.count == 0:
                    self.clearing_price = self.curve_seller.price[0] - self.bid_offset
                elif self.curve_seller.count == 0 and self.curve_buyer.count > 0:
                    self.clearing_price = self.curve_buyer.price[0] + self.bid_offset
                else:
                    if self.curve_seller.price[0] == self.pricecap:
                        self.clearing_price = self.curve_buyer.price[0] + self.bid_offset
                    elif self.curve_seller.price[0] == -self.pricecap:
                        self.clearing_price = self.curve_seller.price[0] - self.bid_offset  
                    else:
                        self.clearing_price = self.curve_seller.price[0] + (self.curve_buyer.price[0] - self.curve_seller.price[0]) * self.clearing_scalar
           
            elif self.clearing_quantity < self.unresponsive_buy:
                self.clearing_type = helpers.ClearingType.FAILURE
                self.clearing_price = self.pricecap
            
            elif self.clearing_quantity < self.unresponsive_sell:
                self.clearing_type = helpers.ClearingType.FAILURE
                self.clearing_price = -self.pricecap
            
            elif self.clearing_quantity == self.unresponsive_buy and self.clearing_quantity == self.unresponsive_sell:
                # only cleared unresponsive loads
                self.clearing_type = helpers.ClearingType.PRICE
                self.clearing_price = 0.0
            
        # If the market mode MD_NONE and at least one side is not given
        else:
            if self.curve_seller.count > 0 and self.curve_buyer.count == 0:
                self.clearing_price = self.curve_seller.price[0] - self.bid_offset
            elif self.curve_seller.count == 0 and self.curve_buyer.count > 0:
                self.clearing_price = self.curve_buyer.price[0] + self.bid_offset
            elif self.curve_seller.count > 0 and self.curve_buyer.count > 0:
                self.clearing_price = self.curve_seller.price[0] + (self.curve_buyer.price[0] - self.curve_seller.price[0]) * self.clearing_scalar
            elif self.curve_seller.count == 0 and self.curve_buyer.count == 0:
                self.clearing_price = 0.0
            self.clearing_quantity = 0
            self.clearing_type = helpers.ClearingType.NULL
            if self.curve_seller.count == 0 :
                missingBidder = "seller"
            elif self.curve_buyer.count == 0:
                missingBidder = "buyer"
            print ('  Market %s fails to clear due to missing %s' % (self.name, missingBidder), flush=True)
            
        # Calculation of the marginal 
        marginal_total = self.marginal_quantity = self.marginal_frac = 0.0
        if self.clearing_type == helpers.ClearingType.BUYER:
            marginal_subtotal = 0
            i = 0
            for i in range(self.curve_buyer.count):
                if self.curve_buyer.price[i] > self.clearing_price:
                    marginal_subtotal = marginal_subtotal + self.curve_buyer.quantity[i]
                else:
                    break
            self.marginal_quantity =  self.clearing_quantity - marginal_subtotal
            for j in range(i, self.curve_buyer.count):
                if self.curve_buyer.price[i] == self.clearing_price:
                    marginal_total += self.curve_buyer.quantity[i]
                else:
                    break
            if marginal_total > 0.0:
                self.marginal_frac = float(self.marginal_quantity) / marginal_total
       
        elif self.clearing_type == helpers.ClearingType.SELLER:
            marginal_subtotal = 0
            i = 0
            for i in range(0, self.curve_seller.count):
                if self.curve_seller.price[i] > self.clearing_price:
                    marginal_subtotal = marginal_subtotal + self.curve_seller.quantity[i]
                else:
                    break
            self.marginal_quantity =  self.clearing_quantity - marginal_subtotal
            for j in range(i, self.curve_seller.count):
                if self.curve_seller.price[i] == self.clearing_price:
                    marginal_total += self.curve_seller.quantity[i]
                else:
                    break
            if marginal_total > 0.0:
                self.marginal_frac = float (self.marginal_quantity) / marginal_total 
        
        else:
            self.marginal_quantity = 0.0
            self.marginal_frac = 0.0
        # print ('##', time_granted, tnext_clear, self.clearing_type, self.clearing_quantity, 
        #        self.clearing_price,
        #        self.curve_buyer.count, self.unresponsive_buy, self.responsive_buy,
        #        self.curve_seller.count, self.unresponsive_sell, self.responsive_sell,
        #        self.marginal_quantity, self.marginal_frac, self.lmp, self.refload,
        #        self.consumerSurplus, self.averageConsumerSurplus, self.supplierSurplus,
        #        self.unrespSupplierSurplus, sep=',', flush=True)

    def surplusCalculation(self, tnext_clear=0, time_granted=0):
        """Calculates consumer surplus (and its average) and supplier surplus.

        This function goes through all the bids higher than clearing price from buyers to calculate consumer surplus,
         and also accumlates the quantities that will be cleared while doing so. Of the cleared quantities,
         the quantity for unresponsive loads are also collected.
         Then go through each seller to calculate supplier surplus.
         Part of the supplier surplus corresponds to unresponsive load are excluded and calculated separately.

        :param tnext_clear (int): next clearing time in FNCS seconds, should be <= time_granted, for the log file only
        :param time_granted (int): the current time in FNCS seconds, for the log file only
        :return: None
        """
        numberOfUnrespBuyerAboveClearingPrice = 0
        numberOfResponsiveBuyerAboveClearingPrice = 0
        self.supplierSurplus = 0.0
        self.averageConsumerSurplus = 0.0
        self.consumerSurplus = 0.0
        self.unrespSupplierSurplus = 0.0
        grantedRespQuantity = 0.0
        grantedUnrespQuantity = 0.0
        declinedQuantity = 0.0
        # assuming the buyers are ordered descending by price
        for i in range(self.curve_buyer.count):
            # if a buyer pays higher than clearing_price, the power is granted
            if self.curve_buyer.price[i] >= self.clearing_price:
                # unresponsive load, they pay infinite amount price, here it is set at self.pricecap
                if self.curve_buyer.price[i] == self.pricecap:
                    grantedUnrespQuantity += self.curve_buyer.quantity[i]
                    numberOfUnrespBuyerAboveClearingPrice += 1
                # responsive load, this is the part consumer surplus is calculated
                else:
                    grantedRespQuantity += self.curve_buyer.quantity[i]
                    numberOfResponsiveBuyerAboveClearingPrice += 1
                    self.consumerSurplus += (self.curve_buyer.price[i] - self.clearing_price) * self.curve_buyer.quantity[i]
            # if a buy pays lower than clearing_price, it does not get the quantity requested
            else:
                declinedQuantity += self.curve_buyer.quantity[i]
        if numberOfResponsiveBuyerAboveClearingPrice != 0:
            self.averageConsumerSurplus = self.consumerSurplus / numberOfResponsiveBuyerAboveClearingPrice
        # assuming the sellers are ordered ascending by their price
        for i in range(self.curve_seller.count):
            # if a seller has a wholesale price/base price lower than clearing_price, their power is used
            if self.curve_seller.price[i] <= self.clearing_price:
                # satisfy quantity requested by unresponsive load first since they pay infinite price
                # when the unresponsive load use up all power from the supplier
                if grantedUnrespQuantity >= self.curve_seller.quantity[i]:
                    self.unrespSupplierSurplus += (self.clearing_price - self.curve_seller.price[i]) * self.curve_seller.quantity[i]
                    grantedUnrespQuantity -= self.curve_seller.quantity[i]
                # when the unresponsive load use part of the power from the supplier
                elif grantedUnrespQuantity != 0.0:
                    self.unrespSupplierSurplus += (self.clearing_price - self.curve_seller.price[i]) * grantedUnrespQuantity
                    leftOverQuantityFromSeller = self.curve_seller.quantity[i] - grantedUnrespQuantity
                    grantedUnrespQuantity = 0.0
                    # leftover quantity from this supplier will be used by responsive load
                    # when leftover quantity is used up by the responsive load
                    if grantedRespQuantity >= leftOverQuantityFromSeller:
                        self.supplierSurplus += (self.clearing_price - self.curve_seller.price[i]) * leftOverQuantityFromSeller
                        grantedRespQuantity -= leftOverQuantityFromSeller
                    # when leftover quantity satisfies all the quantity the responsive load asked
                    else:
                        self.supplierSurplus += (self.clearing_price - self.curve_seller.price[i]) * grantedRespQuantity
                        grantedRespQuantity = 0.0
                        break
                # if the quantity requested by unresponsive load are satisfied, responsive load requests are considered
                # when supplier quantity is used up by the responsive load
                elif grantedRespQuantity >= self.curve_seller.quantity[i]:
                    self.supplierSurplus += (self.clearing_price - self.curve_seller.price[i]) * self.curve_seller.quantity[i]
                    grantedRespQuantity -= self.curve_seller.quantity[i]
                # when supplier quantity satisfies all the quantity the responsive load requested
                else:
                    self.supplierSurplus += (self.clearing_price - self.curve_seller.price[i]) * grantedRespQuantity
                    grantedRespQuantity = 0.0
                    break
        if grantedRespQuantity != 0.0:
            print('cleared {:.4f} more quantity than supplied.'.format(grantedRespQuantity))
        print ('##', 
               time_granted, 
               tnext_clear, 
               self.clearing_type, 
               '{:.3f}'.format(self.clearing_quantity), 
               '{:.6f}'.format(self.clearing_price),
               self.curve_buyer.count, 
               '{:.3f}'.format(self.unresponsive_buy), 
               '{:.3f}'.format(self.responsive_buy),
               self.curve_seller.count, 
               '{:.3f}'.format(self.unresponsive_sell), 
               '{:.3f}'.format(self.responsive_sell),
               '{:.3f}'.format(self.marginal_quantity), 
               '{:.6f}'.format(self.marginal_frac), 
               '{:.6f}'.format(self.lmp), 
               '{:.3f}'.format(self.refload),
               '{:.4f}'.format(self.consumerSurplus), 
               '{:.4f}'.format(self.averageConsumerSurplus), 
               '{:.4f}'.format(self.supplierSurplus),
               '{:.4f}'.format(self.unrespSupplierSurplus), 
               sep=',', flush=True)


# Copyright (C) 2018-2020 Battelle Memorial Institute
# file: precool.py
"""Classes for NIST TE Challenge 2 example

The precool_loop class manages time stepping and FNCS messages
for the precooler agents, which adjust thermostat setpoints in 
response to time-of-use rates and overvoltages. The precooler agents
also estimate house equivalent thermal parameter (ETP) models based
on total floor area, number of stories, number of exterior doors and
and estiamted thermal integrity level. This ETP estimate serves as
an example for other agent developers; it's not actually used by the
precooler agent.

Public Functions:
    :precooler_loop: Initializes and runs the precooler agents.  
"""
import string;
import sys;
try:
  import tesp_support.fncs as fncs;
except:
  pass
import json;
import math;
import re;
if sys.platform != 'win32':
  import resource

def parse_fncs_magnitude (arg):
  """Helper function to find the magnitude of a possibly complex number from FNCS

  Args:
    arg (str): The FNCS value
  """
  tok = arg.strip('+-; MWVAFKdegrij')
  vals = re.split(r'[\+-]+', tok)
  if len(vals) < 2: # only a real part provided
    vals.append('0')
#  vals = [float(v) for v in vals]
#
#  if '-' in tok:
#    vals[1] *= -1.0
  vals[0] = float(vals[0])
  if arg.startswith('-'):
    vals[0] *= -1.0
  return vals[0]

thermalIntegrity = {
  'VERY_LITTLE': {'Rroof':11.0,'Rwall': 4.0,'Rfloor': 4.0,'Rdoors': 3.0,'Rwindows':1/1.27,'airchange_per_hour':1.5},
  'LITTLE':      {'Rroof':19.0,'Rwall':11.0,'Rfloor': 4.0,'Rdoors': 3.0,'Rwindows':1/0.81,'airchange_per_hour':1.5},
  'BELOW_NORMAL':{'Rroof':19.0,'Rwall':11.0,'Rfloor':11.0,'Rdoors': 3.0,'Rwindows':1/0.81,'airchange_per_hour':1.0},
  'NORMAL':      {'Rroof':30.0,'Rwall':11.0,'Rfloor':19.0,'Rdoors': 3.0,'Rwindows':1/0.60,'airchange_per_hour':1.0},
  'ABOVE_NORMAL':{'Rroof':30.0,'Rwall':19.0,'Rfloor':11.0,'Rdoors': 3.0,'Rwindows':1/0.60,'airchange_per_hour':1.0},
  'GOOD':        {'Rroof':30.0,'Rwall':19.0,'Rfloor':22.0,'Rdoors': 5.0,'Rwindows':1/0.47,'airchange_per_hour':0.5},
  'VERY_GOOD':   {'Rroof':48.0,'Rwall':22.0,'Rfloor':30.0,'Rdoors':11.0,'Rwindows':1/0.31,'airchange_per_hour':0.5},
  'UNKNOWN'  :   {'Rroof':30.0,'Rwall':19.0,'Rfloor':22.0,'Rdoors': 5.0,'Rwindows':1/0.47,'airchange_per_hour':0.5}
  }

class precooler:
  """This agent manages the house thermostat for time-of-use and overvoltage responses.

  References:
    `NIST TE Modeling and Simulation Challenge <https://www.nist.gov/engineering-laboratory/smart-grid/hot-topics/transactive-energy-modeling-and-simulation-challenge>`_

  Args:
    name (str): name of this agent
    agentrow (dict): row from the FNCS configuration dictionary for this agent
    gldrow (dict): row from the GridLAB-D metadata dictionary for this agent's house
    k (float): bidding function denominator, in multiples of stddev
    mean (float): mean of the price
    stddev (float): standard deviation of the price
    lockout_time (float): time in seconds between allowed changes due to voltage
    precooling_quiet (float): time of day in seconds when precooling is allowed
    precooling_off (float): time of day in seconds when overvoltage precooling is always turned off

  Attributes:
    name (str): name of this agent
    meterName (str):name of the corresponding triplex_meter in GridLAB-D, from agentrow
    night_set (float): preferred thermostat setpoint during nighttime hours, deg F, from agentrow
    day_set (float): preferred thermostat setpoint during daytime hours, deg F, from agentrow
    day_start_hour (float): hour of the day when daytime thermostat setting period begins, from agentrow
    day_end_hour (float): hour of the day when daytime thermostat setting period ends, from agentrow
    deadband (float): thermostat deadband in deg F, invariant, from agentrow, from agentrow
    vthresh (float): meter line-to-neutral voltage that triggers precooling, from agentrow
    toffset (float): temperature setpoint change for precooling, in deg F, from agentrow
    k (float): bidding function denominator, in multiples of stddev
    mean (float): mean of the price
    stddev (float): standard deviation of the price
    lockout_time (float): time in seconds between allowed changes due to voltage
    precooling_quiet (float): time of day in seconds when precooling is allowed
    precooling_off (float): time of day in seconds when overvoltage precooling is always turned off
    air_temp (float): current air temperature of the house in deg F
    mtr_v (float): current line-neutral voltage at the triplex meter
    basepoint (float): the preferred time-scheduled thermostat setpoint in deg F
    setpoint (float): the thermostat setpoint, including price response, in deg F
    lastchange (float): time of day in seconds when the setpoint was last changed
    precooling (Boolean): True if the house is precooling, False if not
    ti (int): thermal integrity level, as enumerated for GridLAB-D, from gldrow
    sqft (float: total floor area in square feet, from gldrow
    stories (int): number of stories, from gldrow
    doors (int): number of exterior doors, from gldrow
    UA (float): heat loss coefficient
    CA (float): total air thermal mass
    HM (float): interior mass surface conductance
    CM (float): total house thermal mass
  """
  def make_etp_model(self):
    """ Sets the ETP parameters from configuration data

    References:
        `Thermal Integrity Table Inputs and Defaults <http://gridlab-d.shoutwiki.com/wiki/Residential_module_user%27s_guide#Thermal_Integrity_Table_Inputs_and_Defaults>`_
    """
    Rc = thermalIntegrity[self.ti]['Rroof']
    Rw = thermalIntegrity[self.ti]['Rwall']
    Rf = thermalIntegrity[self.ti]['Rfloor']
    Rg = thermalIntegrity[self.ti]['Rwindows']  # g for glazing
    Rd = thermalIntegrity[self.ti]['Rdoors']
    I = thermalIntegrity[self.ti]['airchange_per_hour']
    # some hard-coded GridLAB-D defaults
    aspect = 1.5  # footprint x/y ratio
    A1d = 19.5    # area of one door
    h = 8.0       # ceiling height
    ECR = 1.0     # exterior ceiling fraction
    EFR = 1.0     # exterior floor fraction
    EWR = 1.0     # exterior wall fraction
    WWR = 0.15    # window to exterior wall ratio, 0.07 in the Wiki and 0.15 in GLD
    IWR = 1.5     # interior to exterior wall ratio
    mf = 2.0      # thermal mass per unit floor area, GridLAB-D default is 2 but the TE30 houses range from 3 to almost 5
    hs = 1.46     # interior heat transfer coefficient
    VHa = 0.018

    Ac = (self.sqft / self.stories) * ECR # ceiling area
    Af = (self.sqft / self.stories) * EFR # floor area
    perimeter = 2 * (1 + aspect) * math.sqrt(Ac / aspect) # exterior perimeter
    Awt = self.stories * h * perimeter  # gross exterior wall area
    Ag = WWR * Awt * EWR                # gross window area
    Ad = self.doors * A1d               # total door area
    Aw = (Awt - Ag - Ad) * EWR          # net exterior wall area, taking EWR as 1s
    Vterm = self.sqft * h * VHa

    self.UA = (Ac/Rc) + (Ad/Rd) + (Af/Rf) + (Ag/Rg) + (Aw/Rw) + Vterm*I
    self.CA = 3 * Vterm
    self.HM = hs * (Aw/EWR + Awt*IWR + Ac*self.stories/ECR)
    self.CM = self.sqft * mf - 2*Vterm

    print ('ETP model', self.name, self.ti, '{:.2f}'.format (self.sqft), str(self.stories), str(self.doors))
    print ('  UA', '{:.2f}'.format (self.UA))
    print ('  CA', '{:.2f}'.format (self.CA))
    print ('  HM', '{:.2f}'.format (self.HM))
    print ('  CM', '{:.2f}'.format (self.CM))

  def __init__(self,name,agentrow,gldrow,k,mean,stddev,lockout_time,precooling_quiet,precooling_off,bPrice,bVoltage):
    self.name = name # house name
    self.sqft = gldrow['sqft']
    self.ti = gldrow['thermal_integrity']
    self.stories = gldrow['stories']
    self.doors = gldrow['doors']
    self.meterName = agentrow['meter']
    self.night_set = agentrow['night_set']
    self.day_set = agentrow['day_set']
    self.day_start_hour = agentrow['day_start_hour']
    self.day_end_hour = agentrow['day_end_hour']
    self.deadband = agentrow['deadband']
    self.vthresh = agentrow['vthresh']
    self.toffset = agentrow['toffset']

    # price response
    self.bPrice = bPrice
    self.k = k
    self.mean = mean
    self.stddev = stddev

    # voltage response
    self.bVoltage = bVoltage
    self.lockout_time = lockout_time
    self.precooling_quiet = precooling_quiet
    self.precooling_off = precooling_off
    self.mtr_v = 120.0
    self.air_temp = 78.0
    self.setpoint = 0.0
    self.basepoint = self.night_set
    self.lastchange = -lockout_time
    self.precooling = False

    self.make_etp_model()

  def set_air_temp (self,str):
    """Set the air_temp member variable

    Args:
        str (str): FNCS message with temperature in degrees Fahrenheit
    """
    self.air_temp = parse_fncs_magnitude (str)

  def set_voltage (self,str):
    """ Sets the mtr_v attribute

    Args:
        str (str): FNCS message with meter line-neutral voltage
    """
    self.mtr_v = parse_fncs_magnitude (str)

  def check_setpoint_change (self, hour_of_day, price, time_seconds):
    """Update the setpoint for time of day and price

    Args:
        hour_of_day (float): the current time of day, 0..24
        price (float): the current price in $/kwh
        time_seconds (long long): the current FNCS time in seconds

    Returns:
        Boolean: True if the setpoint changed, False if not
    """
    # time-scheduled changes to the basepoint
    if hour_of_day >= self.day_start_hour and hour_of_day <= self.day_end_hour:
      self.basepoint = self.day_set
    else:
      self.basepoint = self.night_set
    new_setpoint = self.basepoint
    # time-of-day price response
    if self.bPrice:
      tdelta = (price - self.mean) * self.deadband / self.k / self.stddev
      new_setpoint += tdelta
    # overvoltage checks and response
    if self.bVoltage:
      if hour_of_day >= self.precooling_quiet and not self.precooling:
        if self.mtr_v > self.vthresh:
          self.precooling = True
      elif hour_of_day >= self.precooling_off:
        self.precooling = False
      if self.precooling:
        new_setpoint += self.toffset
    # is the new setpoint different from the existing setpoint?
    if abs(new_setpoint - self.setpoint) > 0.1:
      if (time_seconds - self.lastchange) > self.lockout_time:
        self.setpoint = new_setpoint
        self.lastchange = time_seconds
        return True
    return False

  def get_temperature_deviation(self):
    """For metrics, find the difference between air temperature and time-scheduled (preferred) setpoint

    Returns:
        float: absolute value of deviation
    """
    return abs (self.air_temp - self.basepoint)

def precool_loop (nhours, metrics_root, dict_root, response='PriceVoltage'):
  """Function that supervises FNCS messages and time stepping for precooler agents

  Opens metrics_root_agent_dict.json and metrics_root_glm_dict.json for configuration.
  Writes precool_metrics_root.json at completion.

  Args:
    nhours (float): number of hours to simulate
    metrics_root (str): name of the case, without file extension
    dict_root (str): repeat metrics_root, or the name of a shared case dictionary without file extension
    response (str): combination of Price and/or Voltage
  """
  time_stop = int (3600 * nhours)

  lp = open (dict_root + '_agent_dict.json').read()
  dict = json.loads(lp)
  gp = open (dict_root + '_glm_dict.json').read()
  glm_dict = json.loads(gp)

  bPriceResponse = False
  bVoltageResponse = False
  if 'Price' in response:
    bPriceResponse = True
  if 'Voltage' in response:
    bVoltageResponse = True

  precool_meta = {'temperature_deviation_min':{'units':'degF','index':0},
                  'temperature_deviation_max':{'units':'degF','index':1},
                  'temperature_deviation_avg':{'units':'degF','index':2}}
  StartTime = '2013-07-01 00:00:00 PST'
  precool_metrics = {'Metadata':precool_meta,'StartTime':StartTime}

  dt = dict['dt']

  # create and initialize a controller object for each house
  mean = dict['mean']
  stddev = dict['stddev']
  # period = dict['period'] # not used
  k = dict['k_slope']
  precooling_quiet = 4 # disabled before 4 a.m.
  precooling_off = 22 # disabled after 9 p.m.
  lockout_period = 360
  precoolerObjs = {}
  house_keys = list(dict['houses'].keys())
  n_houses = len (house_keys)
  for key in house_keys:
    row = dict['houses'][key]
    gldrow = glm_dict['houses'][key]
    precoolerObjs[key] = precooler (key, row, gldrow, k, mean, stddev, 
                                    lockout_period, precooling_quiet, precooling_off,
                                    bPriceResponse, bVoltageResponse)

  print ('run till', time_stop, 'step', dt, 
         'mean', mean, 'stddev', stddev, 'k_slope', k,
         'lockout_period', lockout_period, 'precooling_quiet', precooling_quiet, 'precooling_off', precooling_off,
         'bVoltageResponse', bVoltageResponse, 'bPriceResponse', bPriceResponse)

  fncs.initialize()
  time_granted = 0
  price = 0.11 # mean
  # time_next = dt
  bSetDeadbands = True
  setPrecoolers = set()

  while time_granted < time_stop:
    time_granted = fncs.time_request(time_stop) # time_next
    hour_of_day = 24.0 * ((float(time_granted) / 86400.0) % 1.0)
    events = fncs.get_events()
    for topic in events:
      value = fncs.get_value(topic)
      if topic == 'price':
        price = float(value)
      else:
        pair = topic.split ('#')
        houseName = pair[0]
        if pair[1] == 'V1':
          precoolerObjs[houseName].set_voltage (value)
        elif pair[1] == 'Tair':
          precoolerObjs[houseName].set_air_temp (value)

    if bSetDeadbands:
      bSetDeadbands = False
      print ('setting thermostat deadbands and heating setpoints at', time_granted)
      for key, obj in precoolerObjs.items():
        fncs.publish (key + '_thermostat_deadband', obj.deadband)
        fncs.publish (key + '_heating_setpoint', 60.0)

    # update all of the house setpoints and collect the temperature deviation metrics
    count_temp_dev = 0
    sum_temp_dev = 0.0
    min_temp_dev = 10000.0
    max_temp_dev = 0.0
    n_changes = 0
    max_changes = 25
    for key, obj in precoolerObjs.items():
      if n_changes < max_changes:
        if obj.check_setpoint_change (hour_of_day, price, time_granted):
          print ('  setting {:s} to {:.3f} at {:d} and {:.2f} volts precooling'.format (key,obj.setpoint,time_granted,obj.mtr_v),obj.precooling)
          n_changes += 1
          fncs.publish (key + '_cooling_setpoint', obj.setpoint)
          if obj.precooling:
            setPrecoolers.add (obj.name)
      temp_dev = obj.get_temperature_deviation()
      count_temp_dev += 1
      if temp_dev < min_temp_dev:
        min_temp_dev = temp_dev
      if temp_dev > max_temp_dev:
        max_temp_dev = temp_dev
      sum_temp_dev += temp_dev

    if n_changes > 0:
      print ('*** {:6.4f} hr, changing {:d} setpoints, {:d} out of {:d} are pre-cooling'.format (hour_of_day, n_changes, len(setPrecoolers), n_houses))

    if count_temp_dev < 1:
      count_temp_dev = 1
      min_temp_dev = 0.0
    precool_metrics[str(time_granted)] = [min_temp_dev,max_temp_dev,sum_temp_dev/count_temp_dev]

    time_next = time_granted + dt

  print (len(setPrecoolers), 'houses participated in precooling')
  print ('writing metrics', flush=True)
  mp = open ('precool_' + metrics_root + '_metrics.json', 'w')
  print (json.dumps(precool_metrics), file=mp)
  mp.close()
  print ('done', flush=True)

  print ('finalizing FNCS', flush=True)
  fncs.finalize()

  if sys.platform != 'win32':
    usage = resource.getrusage(resource.RUSAGE_SELF)
    RESOURCES = [
      ('ru_utime', 'User time'),
      ('ru_stime', 'System time'),
      ('ru_maxrss', 'Max. Resident Set Size'),
      ('ru_ixrss', 'Shared Memory Size'),
      ('ru_idrss', 'Unshared Memory Size'),
      ('ru_isrss', 'Stack Size'),
      ('ru_inblock', 'Block inputs'),
      ('ru_oublock', 'Block outputs')]
    print('Resource usage:')
    for name, desc in RESOURCES:
      print('  {:<25} ({:<10}) = {}'.format(desc, name, getattr(usage, name)))




# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: process_houses.py; focus on HVAC
"""Functions to plot house data from GridLAB-D

Public Functions:
    :process_houses: Reads the data and metadata, then makes the plot.  

"""
import json;
import sys;
import numpy as np;
try:
  import matplotlib as mpl;
  import matplotlib.pyplot as plt;
except:
  pass

def process_houses(nameroot, dictname = ''):
  """ Plots the temperature and HVAC power for every house

  This function reads *substation_nameroot_metrics.json* and
  *house_nameroot_metrics.json* for the data;
  it reads *nameroot_glm_dict.json* for the metadata.
  These must all exist in the current working directory.  
  Makes one graph with 2 subplots:
  
  1. Average air temperature at every house
  2. Average HVAC power at every house  

  Args:
    nameroot (str): name of the TESP case, not necessarily the same as the GLM case, without the extension
    dictname (str): metafile name (with json extension) for a different GLM dictionary, if it's not *nameroot_glm_dict.json*. Defaults to empty.
  """
  # first, read and print a dictionary of all the monitored GridLAB-D objects
  if len (dictname) > 0:
      lp = open (dictname).read()
  else:
      lp = open (nameroot + "_glm_dict.json").read()
  dict = json.loads(lp)
  hse_keys = list(dict['houses'].keys())
  hse_keys.sort()
# print("\nHouse Dictionary:")
# for key in hse_keys:
#   row = dict['houses'][key]
# # print (key, "on", row['billingmeter_id'], "has", row['sqft'], "sqft", row['cooling'], "cooling", row['heating'], "heating", row['wh_gallons'], "gal WH")
#   # row['feeder_id'] is also available

  # Houses 
  lp_h = open ("house_" + nameroot + "_metrics.json").read()
  lst_h = json.loads(lp_h)
  lst_h.pop('StartTime')
  meta_h = lst_h.pop('Metadata')
  times = list(map(int,list(lst_h.keys())))
  times.sort()
  print ("There are", len (times), "sample times at", times[1] - times[0], "second intervals")
  hrs = np.array(times, dtype=np.float)
  denom = 3600.0
  hrs /= denom

#  print("\nHouse Metadata for", len(lst_h[time_key]), "objects")
  for key, val in meta_h.items():
  # print (key, val['index'], val['units'])
    if key == 'air_temperature_avg':
      AIR_AVG_IDX = val['index']
    elif key == 'air_temperature_min':
      AIR_MIN_IDX = val['index']
    elif key == 'air_temperature_max':
      AIR_MAX_IDX = val['index']
    elif key == 'hvac_load_avg':
      HVAC_AVG_IDX = val['index']
    elif key == 'hvac_load_min':
      HVAC_MIN_IDX = val['index']
    elif key == 'hvac_load_max':
      HVAC_MAX_IDX = val['index']
    elif key == 'waterheater_load_avg':
      WH_AVG_IDX = val['index']
    elif key == 'waterheater_load_min':
      WH_MIN_IDX = val['index']
    elif key == 'waterheater_load_max':
      WH_MAX_IDX = val['index']
    elif key == 'total_load_avg':
      TOTAL_AVG_IDX = val['index']
    elif key == 'total_load_min':
      TOTAL_MIN_IDX = val['index']
    elif key == 'total_load_max':
      TOTAL_MAX_IDX = val['index']
    elif key == 'air_temperature_deviation_cooling':
      DEV_COOL_IDX = val['index']
    elif key == 'air_temperature_deviation_heating':
      DEV_HEAT_IDX = val['index']

  time_key = str(times[0])
  data_h = np.empty(shape=(len(hse_keys), len(times), len(lst_h[time_key][hse_keys[0]])), dtype=np.float)
  print ("\nConstructed", data_h.shape, "NumPy array for Houses")
  j = 0
  for key in hse_keys:
    i = 0
    for t in times:
      ary = lst_h[str(t)][hse_keys[j]]
      data_h[j, i,:] = ary
      i = i + 1
    j = j + 1

  # display a plot
  fig, ax = plt.subplots(2, 1, sharex = 'col')
  i = 0
  for key in hse_keys:
    ax[0].plot(hrs, data_h[i,:,AIR_AVG_IDX], color="blue")
    ax[1].plot(hrs, data_h[i,:,HVAC_AVG_IDX], color="red")
    i = i + 1
  ax[0].set_ylabel("Degrees")
  ax[1].set_ylabel("kW")
  ax[1].set_xlabel("Hours")
  ax[0].set_title ("HVAC at all Houses")

  plt.show()





# Copyright (C) 2017-2020 Battelle Memorial Institute
# file: tesp_config.py
"""Presents a GUI to configure and package TESP cases

Public Functions:
  :show_tesp_config: Initializes and runs the GUI

References:
  `Graphical User Interfaces with Tk <https://docs.python.org/3/library/tk.html>`_
"""
import sys
import json
import csv
import tkinter as tk
import tkinter.ttk as ttk
import numpy as np
from tkinter import filedialog
from tkinter import messagebox
import os
import shutil

config = {'BackboneFiles':{},'FeederGenerator':{},'EplusConfiguration':{},'PYPOWERConfiguration':{},
  'AgentPrep':{},'ThermostatSchedule':{},'WeatherPrep':{},'SimulationConfig':{},'MonteCarloCase':{}}

StartTime = "2013-07-01 00:00:00"
EndTime = "2013-07-03 00:00:00"

taxonomyChoices = ['TE_Base',
                   'GC-12.47-1',
                   'R1-12.47-1',
                   'R1-12.47-2',
                   'R1-12.47-3',
                   'R1-12.47-4',
                   'R1-25.00-1',
                   'R2-12.47-1',
                   'R2-12.47-2',
                   'R2-12.47-3',
                   'R2-25.00-1',
                   'R2-35.00-1',
                   'R3-12.47-1',
                   'R3-12.47-2',
                   'R3-12.47-3',
                   'R4-12.47-1',
                   'R4-12.47-2',
                   'R4-25.00-1',
                   'R5-12.47-1',
                   'R5-12.47-2',
                   'R5-12.47-3',
                   'R5-12.47-4',
                   'R5-12.47-5',
                   'R5-25.00-1',
                   'R5-35.00-1'
                   ];

buildingChoices = ['FullServiceRestaurant',
                   'Hospital',
                   'LargeHotel',
                   'LargeOffice',
                   'MediumOffice',
                   'MidriseApartment',
                   'OutPatient',
                   'PrimarySchool',
                   'QuickServiceRestaurant',
                   'SchoolDualController',
                   'SecondarySchool',
                   'SmallHotel',
                   'SmallOffice',
                   'StandaloneRetail',
                   'StripMall',
                   'SuperMarket',
                   'Warehouse'
                   ]

inverterModesBattery = ['GROUP_LOAD_FOLLOWING','LOAD_FOLLOWING','VOLT_VAR_FREQ_PWR','VOLT_WATT','VOLT_VAR','CONSTANT_PF','CONSTANT_PQ','NONE'];
inverterModesPV = ['VOLT_VAR_FREQ_PWR','VOLT_WATT','VOLT_VAR','CONSTANT_PF','CONSTANT_PQ','NONE'];

billingModes = ['TIERED_TOU','TIERED_RTP','HOURLY','TIERED','UNIFORM','NONE'];

eplusVoltageChoices = ['208', '480'];

powerFlowChoices = ['AC','DC'];
optimalPowerFlowChoices = ['AC','DC'];

# if one of these ends in 'Mid' we need to supply a 'Band' input
monteCarloChoices = ['None','ElectricCoolingParticipation','ThermostatRampMid','ThermostatOffsetLimitMid',
                     'WeekdayEveningStartMid','WeekdayEveningSetMid'];

weatherChoices = ['TMY3'] # ,'CSV','WeatherBug']

# var columns are label, value, hint, JSON class, JSON attribute
# if there is a sixth column, that will be the name of a Choices tuple, to be edited via Combobox
# if there is not a sixth column, that indicates a single value to be edited via Entry
varsTM = [['Start Time',StartTime,'GLD Date/Time','SimulationConfig','StartTime'],
          ['End Time',EndTime,'GLD Date/Time','SimulationConfig','EndTime'],
          ['Market Clearing Period',300,'s','AgentPrep','MarketClearingPeriod'],
          ['GridLAB-D Time Step',15,'s','FeederGenerator','MinimumStep'],
          ['Metrics Time Step',300,'s','FeederGenerator','MetricsInterval'],
          ['Power Flow Time Step',15,'s','PYPOWERConfiguration','PFStep'],
          ['Agent Time Step',15,'s','AgentPrep','TimeStepGldAgents'],
          ['GridLAB-D Taxonomy Choice','TE_Base','','BackboneFiles','TaxonomyChoice','taxonomyChoices'],
          ['Feeder Name Prefix','Fdr1_','Prepend to Taxonomy Names','BackboneFiles','NamePrefix'],
          ['PYPOWER Base File','ppbasefile.py','','BackboneFiles','PYPOWERFile'],
          ['Weather Type','TMY3','','WeatherPrep','WeatherChoice','weatherChoices'],
          ['Weather Source','WA-Yakima_Air_Terminal.tmy3','File or URL','WeatherPrep','DataSource'],
          ['Latitude (N > 0)',32.133,'deg','WeatherPrep','Latitude'],
          ['Longitude (E > 0)',-110.95,'deg','WeatherPrep','Longitude'],
          ['Altitude (not used yet)', 777.0,'m','WeatherPrep','Altitude'],
          ['TZmeridian (E > 0)',-105.00,'deg','WeatherPrep','TZmeridian'],
          ['Support Directory','$TESP_INSTALL/share/support','Parent directory of base model files','SimulationConfig','SourceDirectory'], # row 13 for TESPDIR
          ['Working Directory','./','','SimulationConfig','WorkingDirectory'],
          ['Case Name','Test','','SimulationConfig','CaseName']
          ];
varsTMSupportDirIndex = 16

varsFD = [['Electric Cooling Penetration',90,'%','FeederGenerator','ElectricCoolingPercentage'],
          ['Electric Cooling Participation',50,'%','FeederGenerator','ElectricCoolingParticipation'],
          ['Water Heater Penetration',75,'%','FeederGenerator','WaterHeaterPercentage'],
          ['Water Heater Participation',25,'%','FeederGenerator','WaterHeaterParticipation'],
          ['Solar Penetration',0,'%','FeederGenerator','SolarPercentage'],
          ['Storage Penetration',0,'%','FeederGenerator','StoragePercentage'],
          ['Solar Inverter Mode','CONSTANT_PF','','FeederGenerator','SolarInverterMode','inverterModesPV'],
          ['Storage Inverter Mode','CONSTANT_PF','','FeederGenerator','StorageInverterMode','inverterModesBattery'],
          ['Billing Mode','TIERED','','FeederGenerator','BillingMode','billingModes'],
          ['Monthly Fee',13,'$','FeederGenerator','MonthlyFee'],
          ['Price',0.102013,'$/kwh','FeederGenerator','Price'],
          ['Tier 1 Energy',500,'kwh','FeederGenerator','Tier1Energy'],
          ['Tier 1 Price',0.117013,'$/kwh','FeederGenerator','Tier1Price'],
          ['Tier 2 Energy',1000,'kwh','FeederGenerator','Tier2Energy'],
          ['Tier 2 Price',0.122513,'$/kwh','FeederGenerator','Tier2Price'],
          ['Tier 3 Energy',0,'kwh','FeederGenerator','Tier3Energy'],
          ['Tier 3 Price',0,'$/kwh','FeederGenerator','Tier3Price']
          ];
varsPP = [['OPF Type','DC','for dispatch and price','PYPOWERConfiguration','ACOPF','optimalPowerFlowChoices'],
          ['PF Type','DC','for voltage','PYPOWERConfiguration','ACPF','powerFlowChoices'],
          ['Substation Voltage',230.0,'kV','PYPOWERConfiguration','TransmissionVoltage'],
          ['Substation Base',12.0,'MVA','PYPOWERConfiguration','TransformerBase'],
          ['GLD Bus Number',7,'','PYPOWERConfiguration','GLDBus'],
          ['GLD Load Scale',20,'','PYPOWERConfiguration','GLDScale'],
          ['Non-responsive Loads','NonGLDLoad.txt','CSV File','PYPOWERConfiguration','CSVLoadFile'],
          ['Unit Out',2,'','PYPOWERConfiguration','UnitOut'],
          ['Unit Outage Start','2013-07-02 07:00:00','GLD Date/Time','PYPOWERConfiguration','UnitOutStart'],
          ['Unit Outage End','2013-07-02 19:00:00','GLD Date/Time','PYPOWERConfiguration','UnitOutEnd'],
          ['Branch Out',3,'','PYPOWERConfiguration','BranchOut'],
          ['Branch Outage Start','','GLD Date/Time','PYPOWERConfiguration','BranchOutStart'],
          ['Branch Outage End','','GLD Date/Time','PYPOWERConfiguration','BranchOutEnd']
          ];
varsEP = [['Energy+ Base File','SchoolDualController','','EplusConfiguration','BuildingChoice','buildingChoices'],
          ['Energy+ EMS File', 'ems/emsSchoolDualController', '','EplusConfiguration','EMSFile'],
          ['Time Steps per Hour',12,'','EplusConfiguration','StepsPerHour'],
          ['Eplus Bus','Feeder_head','','EplusConfiguration','EnergyPlusBus'],
          ['Eplus Service Voltage',480,'V','EplusConfiguration','EnergyPlusServiceV','eplusVoltageChoices'],
          ['Eplus Transformer Size',150,'kVA','EplusConfiguration','EnergyPlusXfmrKva'],
          ['Reference Price',0.02,'$','EplusConfiguration','ReferencePrice'],
          ['Ramp',25,'degF/$','EplusConfiguration','Slope'],
          ['Delta Limit Hi',4,'degF','EplusConfiguration','OffsetLimitHi'],
          ['Delta Limit Lo',4,'degF','EplusConfiguration','OffsetLimitLo']
          ];
varsAC = [['Initial Price',0.02078,'$','AgentPrep','InitialPriceMean'],
          ['Std Dev Price',0.00361,'$','AgentPrep','InitialPriceStdDev'],
          ['Ramp Lo',0.5,'$(std dev)/degF','AgentPrep','ThermostatRampLo'],
          ['Ramp Hi',3.0,'$(std dev)/degF','AgentPrep','ThermostatRampHi'],
          ['Band Lo',1.0,'degF','AgentPrep','ThermostatBandLo'],
          ['Band Hi',3.0,'degF','AgentPrep','ThermostatBandHi'],
          ['Offset Limit Lo',2.0,'degF','AgentPrep','ThermostatOffsetLimitLo'],
          ['Offset Limit Hi',6.0,'degF','AgentPrep','ThermostatOffsetLimitHi'],
          ['Price Cap Lo',1.00,'$','AgentPrep','PriceCapLo'],
          ['Price Cap Hi',3.00,'$','AgentPrep','PriceCapHi']
          ];
varsTS = [['Weekday Wakeup Start Lo',5.0,'hour of day','ThermostatSchedule','WeekdayWakeStartLo'],
          ['Weekday Wakeup Start Hi',6.5,'hour of day','ThermostatSchedule','WeekdayWakeStartHi'],
          ['Weekday Wakeup Set Lo',78,'degF','ThermostatSchedule','WeekdayWakeSetLo'],
          ['Weekday Wakeup Set Hi',80,'degF','ThermostatSchedule','WeekdayWakeSetHi'],
          ['Weekday Daylight Start Lo',8.0,'hour of day','ThermostatSchedule','WeekdayDaylightStartLo'],
          ['Weekday Daylight Start Hi',9.0,'hour of day','ThermostatSchedule','WeekdayDaylightStartHi'],
          ['Weekday Daylight Set Lo',84,'degF','ThermostatSchedule','WeekdayDaylightSetLo'],
          ['Weekday Daylight Set Hi',86,'degF','ThermostatSchedule','WeekdayDaylightSetHi'],
          ['Weekday Evening Start Lo',17.0,'hour of day','ThermostatSchedule','WeekdayEveningStartLo'],
          ['Weekday Evening Start Hi',18.5,'hour of day','ThermostatSchedule','WeekdayEveningStartHi'],
          ['Weekday Evening Set Lo',78,'degF','ThermostatSchedule','WeekdayEveningSetLo'],
          ['Weekday Evening Set Hi',80,'degF','ThermostatSchedule','WeekdayEveningSetHi'],
          ['Weekday Night Start Lo',22.0,'hour of day','ThermostatSchedule','WeekdayNightStartLo'],
          ['Weekday Night Start Hi',23.5,'hour of day','ThermostatSchedule','WeekdayNightStartHi'],
          ['Weekday Night Set Lo',72,'degF','ThermostatSchedule','WeekdayNightSetLo'],
          ['Weekday Night Set Hi',74,'degF','ThermostatSchedule','WeekdayNightSetHi'],
          ['Weekend Daylight Start Lo',8.0,'hour of day','ThermostatSchedule','WeekendDaylightStartLo'],
          ['Weekend Daylight Start Hi',9.0,'hour of day','ThermostatSchedule','WeekendDaylightStartHi'],
          ['Weekend Daylight Set Lo',76,'degF','ThermostatSchedule','WeekendDaylightSetLo'],
          ['Weekend Daylight Set Hi',84,'degF','ThermostatSchedule','WeekendDaylightSetHi'],
          ['Weekend Night Start Lo',22.0,'hour of day','ThermostatSchedule','WeekendNightStartLo'],
          ['Weekend Night Start Hi',24.0,'hour of day','ThermostatSchedule','WeekendNightStartHi'],
          ['Weekend Night Set Lo',72,'degF','ThermostatSchedule','WeekendNightSetLo'],
          ['Weekend Night Set Hi',74,'degF','ThermostatSchedule','WeekendNightSetHi']
          ];

class TespConfigGUI:
  """Manages a seven-page GUI for case configuration

  The GUI opens and saves a JSON file in the format used by *tesp.tesp_config*

  Todo:
    * Possible data loss if the user updated the number of Monte Carlo cases, but didn't click the Update button before saving the case configuration.

  Attributes:
    nb (Notebook): the top-level GUI with tabbed pages
    f1 (Frame): the page for date/time setup, along with T&D files, weather files and file paths
    f2 (Frame): the page for feeder generator setup
    f3 (Frame): the page for PYPOWER setup
    f4 (Frame): the page for EnergyPlus setup
    f5 (Frame): the page for simple_auction and hvac agent setup
    f6 (Frame): the page for time-scheduled thermostat settings
    f7 (Frame): the page for Monte Carlo setup
  """
  def __init__ (self, master):
    self.nb = ttk.Notebook (master)
    self.nb.pack(fill='both', expand='yes')

    self.f1 = self.AttachFrame ('varsTM', varsTM)
    self.f2 = self.AttachFrame ('varsFD', varsFD)
    self.f3 = self.AttachFrame ('varsPP', varsPP)
    self.f4 = self.AttachFrame ('varsEP', varsEP)
    self.f5 = self.AttachFrame ('varsAC', varsAC)
    self.f6 = self.AttachFrame ('varsTS', varsTS)
    self.f7 = ttk.Frame (self.nb, name='varsMC')

    self.bldg_ems = self.f4.children['eplusconfiguration#emsfile']
    self.bldg_cb = self.f4.children['eplusconfiguration#buildingchoice']
    self.path_ent = self.f1.children['simulationconfig#sourcedirectory']
    self.tmy3_ent = self.f1.children['weatherprep#datasource']
    self.tz_ent = self.f1.children['weatherprep#tzmeridian']
    self.lat_ent = self.f1.children['weatherprep#latitude']
    self.long_ent = self.f1.children['weatherprep#longitude']
    self.alt_ent = self.f1.children['weatherprep#altitude']

    self.bldg_cb.bind("<<ComboboxSelected>>", self.UpdateEMS)

    #ttk.Style().configure('TButton', background='blue')
    ttk.Style().configure('TButton', foreground='blue')
    btn = ttk.Button(self.f1, text='Lat/Long/Alt/TZ from TMY3', command=self.ReadLatLong)
    btn.grid(row=len(varsTM) + 2, column=1, sticky=tk.NSEW)
    btn = ttk.Button(self.f1, text='Save Config...', command=self.SaveConfig)
    btn.grid(row=len(varsTM) + 3, column=1, sticky=tk.NSEW)
    btn = ttk.Button(self.f1, text='Open Config...', command=self.OpenConfig)
    btn.grid(row=len(varsTM) + 4, column=1, sticky=tk.NSEW)

    lab = ttk.Label(self.f7, text='Columns', relief=tk.RIDGE)
    lab.grid(row=0, column=0, sticky=tk.NSEW)
    cb = ttk.Combobox(self.f7, values=monteCarloChoices, name='cb1')
    cb.set(monteCarloChoices[1])
    cb.grid(row=0, column=1, sticky=tk.NSEW)
    cb = ttk.Combobox(self.f7, values=monteCarloChoices, name='cb2')
    cb.set(monteCarloChoices[2])
    cb.grid(row=0, column=2, sticky=tk.NSEW)
    cb = ttk.Combobox(self.f7, values=monteCarloChoices, name='cb3')
    cb.set(monteCarloChoices[3])
    cb.grid(row=0, column=3, sticky=tk.NSEW)

    self.InitializeMonteCarlo (7)
    lab = ttk.Label(self.f7, text='Rows', relief=tk.RIDGE)
    lab.grid(row=1, column=0, sticky=tk.NSEW)
    ent = ttk.Entry(self.f7, name='rows')
    ent.insert(0, config['MonteCarloCase']['NumCases'])
    ent.grid(row=1, column=1, sticky=tk.NSEW)
    btn = ttk.Button (self.f7, text='Update', command=self.UpdateMonteCarloFrame)
    btn.grid(row=1, column=3, sticky=tk.NSEW)
    self.SizeMonteCarlo (config['MonteCarloCase']['NumCases'])
    self.SizeMonteCarloFrame (self.f7)

    self.nb.add(self.f1, text='Main', underline=0, padding=2)
    self.nb.add(self.f2, text='Feeder', underline=0, padding=2)
    self.nb.add(self.f3, text='PYPOWER', underline=0, padding=2)
    self.nb.add(self.f4, text='Energy+', underline=0, padding=2)
    self.nb.add(self.f5, text='Auction', underline=0, padding=2)
    self.nb.add(self.f6, text='Thermostats', underline=0, padding=2)
    self.nb.add(self.f7, text='Sampling', underline=0, padding=2)

  def AttachFrame(self, tag, vars):
    """Creates a GUI page and loads it with data

    Label, Combobox and Entry (i.e. edit) controls are automatically created for each row of data

    Args:
      tag (str): the name of the Frame, i.e., GUI page
      vars (dict): the section of case configuration data to be loaded onto this new GUI page
    """
    f = ttk.Frame(self.nb, name=tag)
    lab = ttk.Label(f, text='Parameter', relief=tk.RIDGE)
    lab.grid(row=0, column=0, sticky=tk.NSEW)
    lab = ttk.Label(f, text='Value', relief=tk.RIDGE)
    lab.grid(row=0, column=1, sticky=tk.NSEW)
    lab = ttk.Label(f, text='Units/Notes', relief=tk.RIDGE)
    lab.grid(row=0, column=2, sticky=tk.NSEW)
    for i in range(len(vars)):
      lab = ttk.Label(f, text=vars[i][0], relief=tk.RIDGE)
      lab.grid(row=i+1, column=0, sticky=tk.NSEW)
      varName = (vars[i][3] + '#' + vars[i][4]).lower()
      if len(vars[i]) > 5:
        cb = ttk.Combobox(f, values=globals()[vars[i][5]], name=varName)
        cb.set(vars[i][1])
        cb.grid(row=i+1, column=1, sticky=tk.NSEW)
      else:
        ent = ttk.Entry(f, name=varName)
        ent.insert(0, vars[i][1])
        ent.grid(row=i+1, column=1, sticky=tk.NSEW)
      lab = ttk.Label(f, text=vars[i][2], relief=tk.RIDGE)
      lab.grid(row=i+1, column=2, sticky=tk.NSEW)
    f.columnconfigure (0, weight=1)
    f.columnconfigure (1, weight=2)
    f.columnconfigure (2, weight=1)
    return f

  def ReloadFrame(self, f, vars):
    """Helper function to recreate the GUI page controls and load them with values

    Args:
      f (Frame): the GUI page to reload
      vars (dict): the section of case configuration with values to be loaded
    """
    for i in range(len(vars)):
      ent = f.grid_slaves (row=i+1, column=1)[0]
      ent.delete (0, tk.END)
      ent.insert (0, vars[i][1])

  def mcSample (self, var):
    """Return an appropriate random value for each Monte Carlo variable choice

    Args:
      var (str): one of ElectricCoolingParticipation, ThermostatRampMid, ThermostatOffsetLimitMid, WeekdayEveningStartMid or WeekdayEveningSetMid
    """
    if var == 'ElectricCoolingParticipation':
      return '{:.3f}'.format(np.random.uniform (0, 100))
    elif var == 'ThermostatRampMid':
      return '{:.3f}'.format(np.random.uniform (1.0, 4.0))
    elif var == 'ThermostatOffsetLimitMid':
      return '{:.3f}'.format(np.random.uniform (0, 6.0))
    elif var == 'WeekdayEveningStartMid':
      return '{:.3f}'.format(np.random.uniform (16.5, 18.0))
    elif var == 'WeekdayEveningSetMid':
      return '{:.3f}'.format(np.random.uniform (68.0, 74.0))
    else:
      return '{:.3f}'.format(np.random.uniform (0, 1))

  def mcBand (self, var):
    """Find the band size corresponding to each Monte Carlo variable choice

    Args:
      var (str): one of ElectricCoolingParticipation, ThermostatRampMid, ThermostatOffsetLimitMid, WeekdayEveningStartMid or WeekdayEveningSetMid
    """
    if var == 'ElectricCoolingParticipation':
      return 10.0
    elif var == 'ThermostatRampMid':
      return 0.5
    elif var == 'ThermostatOffsetLimitMid':
      return 2.0
    elif var == 'WeekdayEveningStartMid':
      return 1.0
    elif var == 'WeekdayEveningSetMid':
      return 1.0
    else:
      return 0.0

  def SizeMonteCarlo(self, n):
    """Initializes the Monte Carlo data structures with variable choices and samples

    Args:
      n (int): the number of Monte Carlo shots
    """
    var1 = config['MonteCarloCase']['Variable1']
    var2 = config['MonteCarloCase']['Variable2']
    var3 = config['MonteCarloCase']['Variable3']
    config['MonteCarloCase']['NumCases'] = n
    config['MonteCarloCase']['Band1'] = self.mcBand (var1)
    config['MonteCarloCase']['Band2'] = self.mcBand (var2)
    config['MonteCarloCase']['Band3'] = self.mcBand (var3)
    config['MonteCarloCase']['Samples1'] = [0] * n
    config['MonteCarloCase']['Samples2'] = [0] * n
    config['MonteCarloCase']['Samples3'] = [0] * n
    for i in range(n):
      config['MonteCarloCase']['Samples1'][i] = self.mcSample (var1) 
      config['MonteCarloCase']['Samples2'][i] = self.mcSample (var2)
      config['MonteCarloCase']['Samples3'][i] = self.mcSample (var3)

  def InitializeMonteCarlo(self, n):
    """Makes default variable choices and then initializes the Monte Carlo GUI page

    Args:
      n (int): the number of Monte Carlo shots
    """
    config['MonteCarloCase']['Variable1'] = monteCarloChoices[1]
    config['MonteCarloCase']['Variable2'] = monteCarloChoices[2]
    config['MonteCarloCase']['Variable3'] = monteCarloChoices[3]
    self.SizeMonteCarlo(n)

  # row 0 for dropdowns, 1 for update controls, 2 for column headers, 3 for range edits
  def SizeMonteCarloFrame(self, f):
    """Update the Monte Carlo page to match the number of shots and variables

    Args:
      f (Frame): the Monte Carlo GUI page
    """
    startRow = 3
    for w in f.grid_slaves():
      if int(w.grid_info()['row']) > 2:
        w.grid_forget()

    col1 = f.children['cb1'].get()
    col2 = f.children['cb2'].get()
    col3 = f.children['cb3'].get()
    use1 = col1 != 'None'
    use2 = col2 != 'None'
    use3 = col3 != 'None'
    band1 = 'Mid' in col1
    band2 = 'Mid' in col2
    band3 = 'Mid' in col3

    lab = ttk.Label(f, text='Case #', relief=tk.RIDGE)
    lab.grid(row=startRow + 1, column=0, sticky=tk.NSEW)
    lab = ttk.Label(f, text=col1, relief=tk.RIDGE)
    lab.grid(row=startRow - 1, column=1, sticky=tk.NSEW)
    lab = ttk.Label(f, text=col2, relief=tk.RIDGE)
    lab.grid(row=startRow - 1, column=2, sticky=tk.NSEW)
    lab = ttk.Label(f, text=col3, relief=tk.RIDGE)
    lab.grid(row=startRow - 1, column=3, sticky=tk.NSEW)

    lab = ttk.Label(f, text='Band', relief=tk.RIDGE)
    lab.grid(row=startRow, column=0, sticky=tk.NSEW)
    if band1:
      w1 = ttk.Entry(f)
      w1.insert(0, config['MonteCarloCase']['Band1'])
    else:
      w1 = ttk.Label(f, text='n/a', relief=tk.RIDGE)
    if band2:
      w2 = ttk.Entry(f)
      w2.insert(0, config['MonteCarloCase']['Band2'])
    else:
      w2 = ttk.Label(f, text='n/a', relief=tk.RIDGE)
    if band3:
      w3 = ttk.Entry(f)
      w3.insert(0, config['MonteCarloCase']['Band3'])
    else:
      w3 = ttk.Label(f, text='n/a', relief=tk.RIDGE)
    w1.grid(row=startRow, column=1, sticky=tk.NSEW)
    w2.grid(row=startRow, column=2, sticky=tk.NSEW)
    w3.grid(row=startRow, column=3, sticky=tk.NSEW)

    n = int (config['MonteCarloCase']['NumCases'])
    for i in range(n):
      lab = ttk.Label(f, text=str(i+1), relief=tk.RIDGE)
      lab.grid(row=i+2+startRow, column=0, sticky=tk.NSEW)
      if use1:
        w1 = ttk.Entry(f)
        w1.insert(0, config['MonteCarloCase']['Samples1'][i])
      else:
        w1 = ttk.Label(f, text='n/a', relief=tk.RIDGE)
      if use2:
        w2 = ttk.Entry(f)
        w2.insert(0, config['MonteCarloCase']['Samples2'][i])
      else:
        w2 = ttk.Label(f, text='n/a', relief=tk.RIDGE)
      if use3:
        w3 = ttk.Entry(f)
        w3.insert(0, config['MonteCarloCase']['Samples3'][i])
      else:
        w3 = ttk.Label(f, text='n/a', relief=tk.RIDGE)
      w1.grid(row=i+2+startRow, column=1, sticky=tk.NSEW)
      w2.grid(row=i+2+startRow, column=2, sticky=tk.NSEW)
      w3.grid(row=i+2+startRow, column=3, sticky=tk.NSEW)

  def ReadFrame(self, f, vars):
    """Helper function that reads values from gridded GUI controls into the local case configuration

    Args:
      f (Frame): the GUI page to read
      vars (dict): the local data structure to update
    """
    for w in f.grid_slaves():
      col = int(w.grid_info()['column'])
      row = int(w.grid_info()['row'])
      if col == 1 and row > 0 and row <= len(vars):
        val = w.get()
        section = vars[row-1][3]
        attribute = vars[row-1][4]
        config[section][attribute] = val

  def UpdateEMS (self, event):
    emsFile = 'ems/ems' + self.bldg_cb.get()
    self.update_entry (self.bldg_ems, emsFile)

  def update_entry (self, ctl, val):
    ctl.delete (0, tk.END)
    ctl.insert (0, val)

  def ReadLatLong(self):
    """Updates the Latitude and Longitude from TMY3 file
    """
    weatherpath = self.path_ent.get()
    weatherfile = self.tmy3_ent.get()
    fname = os.path.expandvars (os.path.expanduser (weatherpath + '/weather/'+ weatherfile))
    if os.path.isfile(fname):
      fd = open (fname, 'r')
      rd = csv.reader (fd, delimiter=',', skipinitialspace=True)
      row = next(rd)
      tmy3source = row[0]
      tmy3station = row[1]
      tmy3state = row[2]
      tmy3tzoffset = float(row[3])
      tmy3latitude = float(row[4])
      tmy3longitude = float(row[5])
      tmy3altitude = float(row[6])
      self.update_entry (self.tz_ent, tmy3tzoffset)
      self.update_entry (self.lat_ent, tmy3latitude)
      self.update_entry (self.long_ent, tmy3longitude)
      self.update_entry (self.alt_ent, tmy3altitude)
      fd.close()
    else:
      print (fname, 'not found')

  def SaveConfig(self):
    """Updates the local case configuration from the GUI, queries user for a file name, and then saves case configuration to that file
    """
    self.ReadFrame(self.f1, varsTM)
    self.ReadFrame(self.f2, varsFD)
    self.ReadFrame(self.f3, varsPP)
    self.ReadFrame(self.f4, varsEP)
    self.ReadFrame(self.f5, varsAC)
    self.ReadFrame(self.f6, varsTS)

    col1 = self.f7.children['cb1'].get()
    col2 = self.f7.children['cb2'].get()
    col3 = self.f7.children['cb3'].get()
    config['MonteCarloCase']['Variable1'] = col1
    config['MonteCarloCase']['Variable2'] = col2
    config['MonteCarloCase']['Variable3'] = col3
    use1 = col1 != 'None'
    use2 = col2 != 'None'
    use3 = col3 != 'None'
    band1 = 'Mid' in col1
    band2 = 'Mid' in col2
    band3 = 'Mid' in col3
    numCases = int (self.f7.children['rows'].get()) # what if user changed entry and didn't click Update...global numCases?
    for w in self.f7.grid_slaves():
      row = int(w.grid_info()['row'])
      col = int(w.grid_info()['column'])
      if row == 3:
        if col == 1 and band1:
          val = float(w.get())
          config['MonteCarloCase']['Band1'] = val
        if col == 2 and band2:
          val = float(w.get())
          config['MonteCarloCase']['Band2'] = val
        if col == 3 and band3:
          val = float(w.get())
          config['MonteCarloCase']['Band3'] = val
      elif row > 4:
        if col == 1 and use1:
          val = float(w.get())
          config['MonteCarloCase']['Samples1'][row-5] = val
        if col == 2 and use2:
          val = float(w.get())
          config['MonteCarloCase']['Samples2'][row-5] = val
        if col == 3 and use3:
          val = float(w.get())
          config['MonteCarloCase']['Samples3'][row-5] = val
    support_path = os.path.expandvars (os.path.expanduser (config['SimulationConfig']['SourceDirectory']))
    if not os.path.exists (support_path):
      if not messagebox.askyesno ('Continue to Save?', 'TESP Support Directory: ' + support_path + ' not found.'):
        return
    fname = filedialog.asksaveasfilename(initialdir = '~/src/examples/te30',
                                         title = 'Save JSON Configuration to',
                                         defaultextension = 'json')
    if len(fname) > 0:
      op = open (fname, 'w')
      json.dump (config, op, ensure_ascii=False, indent=2)
      op.close()

  def JsonToSection(self, jsn, vars):
    """Helper function that transfers a JSON file segment into GUI data structures

    Args:
      jsn (dict): the loaded JSON file
      vars (dict): the local data structure
    """
    for i in range(len(vars)):
      section = vars[i][3]
      attribute = vars[i][4]
      vars[i][1] = jsn[section][attribute]
      config[section][attribute] = jsn[section][attribute]

  def OpenConfig(self):
    """Opens a JSON case configuration; transfers its data to the GUI
    """
    fname = filedialog.askopenfilename(initialdir = '~/src/examples/te30',
                                      title = 'Open JSON Configuration',
                                      filetypes = (("JSON files","*.json"),("all files","*.*")),
                                      defaultextension = 'json')
    lp = open (fname)
    cfg = json.loads(lp.read())
    lp.close()
    self.JsonToSection (cfg, varsTM)
    self.JsonToSection (cfg, varsFD)
    self.JsonToSection (cfg, varsPP)
    self.JsonToSection (cfg, varsEP)
    self.JsonToSection (cfg, varsAC)
    self.JsonToSection (cfg, varsTS)
    self.ReloadFrame (self.f1, varsTM)
    self.ReloadFrame (self.f2, varsFD)
    self.ReloadFrame (self.f3, varsPP)
    self.ReloadFrame (self.f4, varsEP)
    self.ReloadFrame (self.f5, varsAC)
    self.ReloadFrame (self.f6, varsTS)

    numCases = int (cfg['MonteCarloCase']['NumCases'])
    config['MonteCarloCase']['NumCases'] = numCases
    ent = self.f7.children['rows']
    ent.delete (0, tk.END)
    ent.insert (0, str(numCases))

    var1 = cfg['MonteCarloCase']['Variable1']
    config['MonteCarloCase']['Variable1'] = var1
    self.f7.children['cb1'].set(var1)

    var2 = cfg['MonteCarloCase']['Variable2']
    config['MonteCarloCase']['Variable2'] = var2
    self.f7.children['cb2'].set(var2)

    var3 = cfg['MonteCarloCase']['Variable3']
    config['MonteCarloCase']['Variable3'] = var3
    self.f7.children['cb3'].set(var3)

    config['MonteCarloCase']['Band1'] = cfg['MonteCarloCase']['Band1']
    config['MonteCarloCase']['Band2'] = cfg['MonteCarloCase']['Band2']
    config['MonteCarloCase']['Band3'] = cfg['MonteCarloCase']['Band3']

    config['MonteCarloCase']['Samples1'] = [0] * numCases
    config['MonteCarloCase']['Samples2'] = [0] * numCases
    config['MonteCarloCase']['Samples3'] = [0] * numCases

    for i in range(numCases):
      config['MonteCarloCase']['Samples1'][i] = cfg['MonteCarloCase']['Samples1'][i]
      config['MonteCarloCase']['Samples2'][i] = cfg['MonteCarloCase']['Samples2'][i]
      config['MonteCarloCase']['Samples3'][i] = cfg['MonteCarloCase']['Samples3'][i]

    self.SizeMonteCarloFrame (self.f7)

  def UpdateMonteCarloFrame(self):
    """Transfer data from the Monte Carlo page into the case configuration
    """
    numCases = int (self.f7.children['rows'].get())
    config['MonteCarloCase']['Variable1'] = self.f7.children['cb1'].get()
    config['MonteCarloCase']['Variable2'] = self.f7.children['cb2'].get()
    config['MonteCarloCase']['Variable3'] = self.f7.children['cb3'].get()
    self.SizeMonteCarlo (numCases)
    self.SizeMonteCarloFrame (self.f7)

def show_tesp_config ():
  """Runs the GUI. Reads and writes JSON case configuration files.
  """
  root = tk.Tk()
  root.title('Transactive Energy Simulation Platform: Case Configuration')
  if 'tespdir' in os.environ:
    tespdir = os.environ['tespdir']
    if len(tespdir) > 0:
  #   config['SimulationConfig']['SourceDirectory'] = tespdir
      if sys.platform == 'win32':
        varsTM[varsTMSupportDirIndex][1] = tespdir + '\support'
      else:
        varsTM[varsTMSupportDirIndex][1] = tespdir + '/support'
  my_gui = TespConfigGUI (root)
  while True:
    try:
      root.mainloop()
      break
    except UnicodeDecodeError:
      pass

#show_tesp_config ()



# Copyright (C) 2017-2020 Battelle Memorial Institute
# file: helpers.py
""" Utility functions for use within tesp_support, including new agents.
"""
import numpy as np
import math
import warnings
import re
import sys
from copy import deepcopy
from enum import IntEnum
try:
  import helics
except:
  pass

def idf_int(val):
  """Helper function to format integers for the EnergyPlus IDF input data file

  Args:
    val (int): the integer to format

  Returns:
     str: the integer in string format, padded with a comma and zero or one blanks, in order to fill three spaces
  """
  sval = str(val)
  if len(sval) < 2:
    return sval + ', '
  return sval + ','

def stop_helics_federate (fed):
  print ('finalizing HELICS', flush=True)
  status = helics.helicsFederateFinalize(fed)
  state = helics.helicsFederateGetState(fed)
  assert state == 3
  while helics.helicsBrokerIsConnected(None):
    time.sleep(1)
  helics.helicsFederateFree(fed)
  helics.helicsCloseLibrary()

def zoneMeterName(ldname):
  """ Enforces the meter naming convention for commercial zones

  Commercial zones must be children of load objects. This routine
  replaces "_load_" with "_meter".

  Args:
      objname (str): the GridLAB-D name of a load, ends with _load_##

  Returns:
    str: The GridLAB-D name of upstream meter
  """
  return ldname.replace ('_load_', '_meter_')

# GridLAB-D name should not begin with a number, or contain '-' for FNCS
def gld_strict_name(val):
    """Sanitizes a name for GridLAB-D publication to FNCS

    Args:
        val (str): the input name

    Returns:
        str: val with all '-' replaced by '_', and any leading digit replaced by 'gld\_'
    """
    if val[0].isdigit():
        val = 'gld_' + val
    return val.replace ('-', '_')

class ClearingType (IntEnum):
    """ Describes the market clearing type
    """
    NULL = 0
    FAILURE = 1
    PRICE = 2
    EXACT = 3
    SELLER = 4
    BUYER = 5

class curve:
    """ Accumulates a set of price, quantity bids for later aggregation

    The default order is descending by price.

    Attributes:
        price ([float]): array of prices, in $/kWh
        quantity ([float]): array of quantities, in kW
        count (int): the number of collected bids
        total (float): the total kW bidding
        total_on (float): the total kW bidding that are currently on
        total_off (float): the total kW bidding that are currently off
    """
    def __init__(self):
        self.price = []
        self.quantity = []
        self.count = 0
        self.total = 0.0
        self.total_on = 0.0
        self.total_off = 0.0  

    def set_curve_order(self, flag):
        """ Set the curve order (by price) to ascending or descending

        Args:
            flag (str): 'ascending' or 'descending'
        """
        if flag == 'ascending':
            self.price.reverse()
            self.quantity.reverse()

    def add_to_curve(self, price, quantity, is_on):
        """ Add one point to the curve

        Args:
            price (float): the bid price, should be $/kWhr
            quantity (float): the bid quantity, should be kW
            is_on (Boolean): True if the load is currently on, False if not
        """
        if quantity == 0:
            return
        self.total += quantity
        if is_on:
            self.total_on += quantity
        else:
            self.total_off += quantity
        value_insert_flag = 0
        if self.count == 0:
            # Since it is the first time assigning values to the curve, define an empty array for the price and mean
            self.price = []
            self.quantity = []
            self.price.append(price)
            self.quantity.append(quantity)
            self.count += 1
        else:
            value_insert_flag = 0
            for i in range(0, self.count):
                # If the price is larger than the compared curve section price, price inserted before that section of the curve
                if price >= self.price[i]:
                    if i == 0:
                        # If the price is larger than that of all the curve sections, insert at the beginning of the curve
                        self.price.insert(0, price)
                        self.quantity.insert(0, quantity)
                    else:
                        self.price.insert(i, price)
                        self.quantity.insert(i, quantity)
                    self.count += 1
                    value_insert_flag = 1
                    break

            # If the price is smaller than that of all the curve sections, insert at the end of the curve
            if value_insert_flag == 0:                   
                self.price.append(price)
                self.quantity.append(quantity)
                self.count += 1

def parse_fncs_number (arg):
    """ Parse floating-point number from a FNCS message; must not have leading sign or exponential notation

    Args:
        arg (str): the FNCS string value

    Returns:
        float: the parsed number
    """
    return float(''.join(ele for ele in arg if ele.isdigit() or ele == '.'))

# strip out extra white space, units (deg, degF, V, MW, MVA, KW, KVA) and ;
def parse_fncs_magnitude (arg):
    """ Parse the magnitude of a possibly complex number from FNCS

    Args:
        arg (str): the FNCS string value

    Returns:
        float: the parsed number, or 0 if parsing fails
    """
    try:
        if ('d ' in arg) or ('r ' in arg):  # polar form
            tok = arg.strip('; MWVAKdrij')
            nsign = nexp = ndot = 0
            for i in range(len(tok)):
                if (tok[i] == '+') or (tok[i] == '-'):
                    nsign += 1
                elif (tok[i] == 'e') or (tok[i] == 'E'):
                    nexp += 1
                elif tok[i] == '.':
                    ndot += 1
                if nsign == 1:
                    kpos = i
                if nsign == 2 and nexp == 0:
                    kpos = i
                    break
                if nsign == 3:
                    kpos = i
                    break
            vals = [tok[:kpos],tok[kpos:]]
            vals = [float(v) for v in vals]
            return vals[0]
        tok = arg.strip('; MWVAFKdegri').replace(" ", "") # rectangular form, including real only
        b = complex(tok)
        return abs (b) # b.real
    except:
        print ('parse_fncs_magnitude does not understand', arg)
        return 0

def parse_kw(arg):
    """ Parse the kilowatt load of a possibly complex number from FNCS

    Args:
        arg (str): the FNCS string value

    Returns:
        float: the parsed number in kW, or 0 if parsing fails
    """
    try:
        tok = arg.strip('; MWVAKdrij')
        nsign = nexp = ndot = 0
        for i in range(len(tok)):
            if (tok[i] == '+') or (tok[i] == '-'):
                nsign += 1
            elif (tok[i] == 'e') or (tok[i] == 'E'):
                nexp += 1
            elif tok[i] == '.':
                ndot += 1
            if nsign == 2 and nexp == 0:
                kpos = i
                break
            if nsign == 3:
                kpos = i
                break

        vals = [tok[:kpos],tok[kpos:]]
        vals = [float(v) for v in vals]

        if 'd' in arg:
            vals[1] *= (math.pi / 180.0)
            p = vals[0] * math.cos(vals[1])
            q = vals[0] * math.sin(vals[1])
        elif 'r' in arg:
            p = vals[0] * math.cos(vals[1])
            q = vals[0] * math.sin(vals[1])
        else:
            p = vals[0]
            q = vals[1]

        if 'KVA' in arg:
            p *= 1.0
            q *= 1.0
        elif 'MVA' in arg:
            p *= 1000.0
            q *= 1000.0
        else:  # VA
            p /= 1000.0
            q /= 1000.0

        return p
    except:
        print ('parse_kw does not understand', arg)
        return 0

def aggregate_bid (crv):
    """aggregates the buyer curve into a quadratic or straight-line fit with zero intercept

    Args:
        crv (curve): the accumulated buyer bids

    Returns:
        [float, float, int, float, float]: Qunresp, Qmaxresp, degree, c2 and c1 scaled to MW instead of kW. c0 is always zero.
    """
    unresp = 0
    idx = 0
    pInd = np.flip(np.argsort(np.array(crv.price)), 0)
    p = 1000.0 * np.array (crv.price)[pInd]  # $/MW
    q = 0.001 * np.array (crv.quantity)[pInd] # MWhr
    if p.size > 0:
        idx = np.argwhere (p == p[0])[-1][0]
        unresp = np.cumsum(q[:idx+1])[-1]
    c2 = 0
    c1 = 0
    deg = 0
    n = p.size - idx - 1

    if n < 1:
        qmax = 0
        deg = 0
    else:
        qresp = np.cumsum(q[idx+1:])
        presp = p[idx+1:]
        qmax = qresp[-1]
        cost = np.cumsum(np.multiply(presp, q[idx+1:]))
        if n <= 2:
            A = np.vstack([qresp, np.ones(len(qresp))]).T
            ret = np.linalg.lstsq(A[:, :-1],cost)[0]
            c1 = ret[0]
            deg = 1
        else:
            A = np.vstack([qresp**2, qresp, np.ones(len(qresp))]).T
            ret = np.linalg.lstsq(A[:, :-1],cost,rcond=None)[0]
            c2 = ret[0]
            c1 = ret[1]
            deg = 2
    bid = [unresp, qmax, deg, c2, c1]
    return bid




# Copyright (C) 2017-2020 Battelle Memorial Institute
# file: glm_dict.py
# tuned to feederGenerator_TSP.m for sequencing of objects and attributes
"""Functions to create metadata from a GridLAB-D input (GLM) file

Metadata is written to a JSON file, for convenient loading into a Python 
dictionary.  It can be used for agent configuration, e.g., to initialize a 
forecasting model based on some nominal data.  It's also used with metrics 
output in post-processing.
  
Public Functions:
    :glm_dict: Writes the JSON metadata file.  

"""

import json;
import sys;
import math;
import tesp_support.helpers as helpers;

def ercotMeterName(objname):
  """ Enforces the meter naming convention for ERCOT

  Replaces anything after the last _ with *mtr*.

  Args:
      objname (str): the GridLAB-D name of a house or inverter

  Returns:
    str: The GridLAB-D name of upstream meter
  """
  k = objname.rfind('_')
  root1 = objname[:k]
  k = root1.rfind('_')
  return root1[:k] + '_mtr'

def isCommercialHouse(house_class):
  if ('BIGBOX' in house_class) or ('OFFICE' in house_class) or ('STRIPMALL' in house_class):
    return True
  return False

def ti_enumeration_string(tok):
  """ if thermal_integrity_level is an integer, convert to a string for the metadata
  """
  if tok == '0':
    return 'VERY_LITTLE'
  if tok == '1':
    return 'LITTLE'
  if tok == '2':
    return 'BELOW_NORMAL'
  if tok == '3':
    return 'NORMAL'
  if tok == '4':
    return 'ABOVE_NORMAL'
  if tok == '5':
    return 'GOOD'
  if tok == '6':
    return 'VERY_GOOD'
  if tok == '7':
    return 'UNKNOWN'
  return tok

def glm_dict (nameroot, ercot=False, te30=False):
  """ Writes the JSON metadata file from a GLM file

  This function reads *nameroot.glm* and writes *nameroot_glm_dict.json* 
  The GLM file should have some meters and triplex_meters with the
  bill_mode attribute defined, which identifies them as billing meters
  that parent houses and inverters. If this is not the case, ERCOT naming
  rules can be applied to identify billing meters.

  Args:
      nameroot (str): path and file name of the GLM file, without the extension
      ercot (boolean): request ERCOT billing meter naming. Defaults to false.
      te30 (boolean): request hierarchical meter handling in the 30-house test harness. Defaults to false.
  """
  ip = open (nameroot + '.glm', 'r')
  op = open (nameroot + '_glm_dict.json', 'w')

  FedName = 'gld1'
  feeder_id = 'feeder'
  name = ''
  bulkpowerBus = 'TBD'
  base_feeder = ''
  substationTransformerMVA = 12
  houses = {}
  waterheaters = {}
  billingmeters = {}
  inverters = {}
  feeders = {}
  capacitors = {}
  regulators = {}
  loads = {}

  inSwing = False
  for line in ip:
    lst = line.split()
    if len(lst) > 1:
      if lst[1] == 'substation':
        inSwing = True
      if inSwing == True:
        if lst[0] == 'name':
          feeder_id = lst[1].strip(';')
        if lst[0] == 'groupid':
          base_feeder = lst[1].strip(';')
        if lst[0] == 'base_power':
          substationTransformerMVA = float(lst[1].strip(' ').strip('MVA;')) * 1.0e-6
          if 'MVA' in line:
            substationTransformerMVA *= 1.0e6
          elif 'KVA' in line:
            substationTransformerMVA *= 1.0e3
          inSwing = False
          break

  ip.seek(0,0)
  inHouses = False
  inWaterHeaters = False
  inTriplexMeters = False
  inMeters = False
  inInverters = False
  hasBattery = False
  hasSolar = False
  inCapacitors = False
  inRegulators = False
  inLoads = False
  inFNCSmsg = False
  inHELICSmsg = False
  for line in ip:
    lst = line.split()
    if len(lst) > 1: # terminates with a } or };
      if lst[1] == 'helics_msg':
        inHELICSmsg = True
      if lst[1] == 'fncs_msg':
        inFNCSmsg = True
      if lst[1] == 'house':
        inHouses = True
        parent = ''
        sqft = 2500.0
        cooling = 'NONE'
        heating = 'NONE'
        stories = 1
        thermal_integrity = 'UNKNOWN'
        doors = 4
        house_class = 'SINGLE_FAMILY'
      if inHELICSmsg == True:
        if lst[0] == 'name':
          FedName = lst[1].strip(';')
          inHELICSmsg = False
      if inFNCSmsg == True:
        if lst[0] == 'name':
          FedName = lst[1].strip(';')
          inFNCSmsg = False
      if lst[1] == 'triplex_meter':
        inTriplexMeters = True
        vln = 120.0
        vll = 240.0
        phases = ''
      if lst[1] == 'meter':
        inMeters = True
        vln = 120.0
        vll = 240.0
        phases = 'ABC'
      if lst[1] == 'inverter':
        inInverters = True
        hasBattery = False
        hasSolar = False
        lastInverter = ''
        rating = 25000.0
        inv_eta = 0.9
        bat_eta = 0.8  # defaults without internal battery model
        soc = 1.0
        capacity = 300150.0  # 6 hr * 115 V * 435 A
      if lst[1] == 'capacitor':
        inCapacitors = True
      if lst[1] == 'regulator':
        inRegulators = True
      if lst[1] == 'waterheater':
        inWaterHeaters = True
      if lst[1] == 'load':
        inLoads = True
        parent = ''
      if inLoads == True:
        if lst[0] == 'name':
          name = lst[1].strip(';')
        if lst[0] == 'parent':
          parent = lst[1].strip(';')
          loads[name] = {'parent':parent}
          inLoads = False
      if inCapacitors == True:
        if lst[0] == 'name':
          lastCapacitor = lst[1].strip(';')
          capacitors[lastCapacitor] = {'feeder_id':feeder_id}
          inCapacitors = False
      if inRegulators == True:
        if lst[0] == 'name':
          lastRegulator = lst[1].strip(';')
          regulators[lastRegulator] = {'feeder_id':feeder_id}
          inRegulators = False
      if inInverters == True:
        if lst[0] == 'name' and lastInverter == '':
          lastInverter = lst[1].strip(';')
        if lst[1] == 'solar':
          hasSolar = True
          hasBattery = False
        elif lst[1] == 'battery':
          hasSolar = False
          hasBattery = True
        if lst[0] == 'rated_power':
          rating = float(lst[1].strip(' ').strip(';')) * 1.0
        if lst[0] == 'inverter_efficiency':
          inv_eta = float(lst[1].strip(' ').strip(';')) * 1.0
        if lst[0] == 'round_trip_efficiency':
          bat_eta = float(lst[1].strip(' ').strip(';')) * 1.0
        if lst[0] == 'state_of_charge':
          soc = float(lst[1].strip(' ').strip(';')) * 1.0
        if lst[0] == 'battery_capacity':
          capacity = float(lst[1].strip(' ').strip(';')) * 1.0
      if inHouses == True:
        if lst[0] == 'name':
          name = lst[1].strip(';')
        if lst[0] == 'parent':
          parent = lst[1].strip(';')
        if lst[0] == 'floor_area':
          sqft = float(lst[1].strip(' ').strip(';')) * 1.0
        if lst[0] == 'number_of_doors':
          doors = int(lst[1].strip(' ').strip(';'))
        if lst[0] == 'number_of_stories':
          stories = int(lst[1].strip(' ').strip(';'))
        if lst[0] == 'cooling_system_type':
          cooling = lst[1].strip(';')
        if lst[0] == 'heating_system_type':
          heating = lst[1].strip(';')
        if lst[0] == '//' and lst[1] == 'thermal_integrity_level':
          thermal_integrity = ti_enumeration_string (lst[2].strip(';'))
        if lst[0] == 'groupid':
          house_class = lst[1].strip(';')
        if (lst[0] == 'cooling_setpoint') or (lst[0] == 'heating_setpoint'):
          if ercot:
            lastBillingMeter = ercotMeterName (name)
          elif isCommercialHouse (house_class):
            lastBillingMeter = helpers.zoneMeterName (parent)
          houses[name] = {'feeder_id':feeder_id,'billingmeter_id':lastBillingMeter,'sqft':sqft,'stories':stories,'doors':doors,
            'thermal_integrity':thermal_integrity,'cooling':cooling,'heating':heating,'wh_gallons':0,'house_class':house_class,
            'parent':parent}
          lastHouse = name
          inHouses = False
      if inWaterHeaters == True:
        if lst[0] == 'name': # non-controlled WH won't have a name
          whname = lst[1].strip(' ').strip(';')
          waterheaters[lastHouse] = {'name': whname, 'gallons':0.0, 'tmix': 0.0, 'mlayer': False}
        if lst[0] == 'tank_volume':
          wh_gallons = float(lst[1].strip(' ').strip(';')) * 1.0
          if lastHouse in waterheaters:
            waterheaters[lastHouse]['gallons'] = wh_gallons
          else:
            waterheaters[lastHouse] = {'name': '', 'gallons':wh_gallons, 'tmix': 0.0, 'mlayer': False}
        if lst[0] == 'T_mixing_valve':
          waterheaters[lastHouse]['tmix'] = float(lst[1].strip(' ').strip(';')) * 1.0
        if lst[0] == 'waterheater_model':
          if 'MULTILAYER' == lst[1].strip(' ').strip(';'):
            waterheaters[lastHouse]['mlayer'] = True
      if inTriplexMeters == True:
        if lst[0] == 'name':
          name = lst[1].strip(';')
        if lst[0] == 'phases':
          phases = lst[1].strip(';')
        if lst[0] == 'parent':
          lastMeterParent = lst[1].strip(';')
        if lst[0] == 'bill_mode':
          if te30 == True:
            if 'flatrate' not in name:
              billingmeters[name] = {'feeder_id':feeder_id,'phases':phases,'vll':vll,'vln':vln,'children':[]}
              lastBillingMeter = name
          else:
            billingmeters[name] = {'feeder_id':feeder_id,'phases':phases,'vll':vll,'vln':vln,'children':[]}
            lastBillingMeter = name
          inTriplexMeters = False
      if inMeters == True:
        if lst[0] == 'name':
          name = lst[1].strip(';')
        if lst[0] == 'phases':
          phases = lst[1].strip(';')
        if lst[0] == 'parent':
          lastMeterParent = lst[1].strip(';')
        if lst[0] == 'nominal_voltage':
          vln = float(lst[1].strip(' ').strip(';')) * 1.0
          vll = vln * math.sqrt(3.0)
        if lst[0] == 'bill_mode':
          billingmeters[name] = {'feeder_id':feeder_id,'phases':phases,'vll':vll,'vln':vln,'children':[]}
          lastBillingMeter = name
          inMeters = False
    elif len(lst) == 1:
      if hasSolar:
        if ercot:
          lastBillingMeter = ercotMeterName (name)
        elif te30:
          lastBillingMeter = lastMeterParent
        inverters[lastInverter] = {'feeder_id':feeder_id,
                                  'billingmeter_id':lastBillingMeter,
                                  'rated_W':rating,
                                  'resource':'solar',
                                  'inv_eta':inv_eta}
      elif hasBattery:
        if ercot:
          lastBillingMeter = ercotMeterName (name)
        elif te30:
          lastBillingMeter = lastMeterParent
        inverters[lastInverter] = {'feeder_id':feeder_id,
                                  'billingmeter_id':lastBillingMeter,
                                  'rated_W':rating,
                                  'resource':'battery',
                                  'inv_eta':inv_eta,
                                  'bat_eta':bat_eta,
                                  'bat_capacity':capacity,
                                  'bat_soc':soc}
      hasSolar = False
      hasBattery = False
      inHouses = False
      inWaterHeaters = False
      inTriplexMeters = False
      inMeters = False
      inInverters = False
      inCapacitors = False
      inRegulators = False
      inFNCSmsg = False
      inHELICSmsg = False

  for key, val in houses.items():
    if key in waterheaters:
      val['wh_name'] = waterheaters[key]['name']
      val['wh_gallons'] = waterheaters[key]['gallons']
      val['wh_tmix'] = waterheaters[key]['tmix']
      val['wh_mlayer'] = waterheaters[key]['mlayer']
    if ercot and isCommercialHouse (val['house_class']):
      parent = val['parent']
      val['billingmeter_id'] = loads[parent]['parent']
    mtr = billingmeters[val['billingmeter_id']]
    mtr['children'].append(key)

  for key, val in inverters.items():
    mtr = billingmeters[val['billingmeter_id']]
    mtr['children'].append(key)

  feeders[feeder_id] = {'house_count': len(houses),'inverter_count': len(inverters),'base_feeder':base_feeder}
  substation = {'bulkpower_bus':bulkpowerBus,'FedName':FedName,
    'transformer_MVA':substationTransformerMVA,'feeders':feeders, 
    'billingmeters':billingmeters,'houses':houses,'inverters':inverters,
    'capacitors':capacitors,'regulators':regulators}
  json.dump (substation, op, ensure_ascii=False, indent=2)

  ip.close()
  op.close()
